# -*- coding: utf-8 -*-
# vim: expandtab shiftwidth=4 softtabstop=4
#
import unittest
from unittest_data_provider import data_provider
import os
import shutil
import nextcloud_client as nextcloud
import datetime
import time
import tempfile
import random
import six

from config import Config

def getSupportedDavVersion():
    # connect just to check supported DAV version
    client = nextcloud.Client(Config['owncloud_url'])
    client.login(Config['owncloud_login'], Config['owncloud_password'])

    caps = client.get_capabilities()
    dav_version = None

    if 'dav' in caps and 'chunking' in caps['dav']:
        dav_version = float(caps['dav']['chunking'])

    return dav_version

def skipIfDavVersionLessThan(expected_version):
    dav_version = getSupportedDavVersion()
    if dav_version is None or dav_version < expected_version:
        return unittest.skip("Expected DAV version %s not supported in this ownCloud version" % expected_version)
    return lambda func: func

class TestFileAccess(unittest.TestCase):

    def get_dav_endpoint_version(self):
        return 0

    def files():
        return (
            ['test.txt'],
            ['test space and + and #.txt'],
            [u'文件.txt']
        )

    def files_content():
        return (
            ['test.txt', b'Hello world!', 'subdir'],
            ['test space and + and #.txt', b'Hello space with+plus#hash!', 'subdir with space + plus and #hash'],
            [u'文件.txt', u'你好世界'.encode('utf-8'), u'文件夹']
        )

    def setUp(self):
        self.temp_dir = tempfile.gettempdir() + '/pyocclient_test%s-%s/' % (int(time.time()), random.randint(1, 1000))
        os.mkdir(self.temp_dir)

        self.client = nextcloud.Client(Config['owncloud_url'], dav_endpoint_version=self.get_dav_endpoint_version())
        self.client.login(Config['owncloud_login'], Config['owncloud_password'])
        self.test_root = Config['test_root']
        if not self.test_root[-1] == '/':
            self.test_root += '/'
        if not self.test_root[0] == '/':
            self.test_root = '/' + self.test_root
        self.client.mkdir(self.test_root)
        self.share2user = Config['owncloud_share2user']
        self.share2userPwd = 'Avcpwd4l!'
        self.test_group = Config['test_group']
        try:
            self.client.create_user(self.share2user, self.share2userPwd)
        except:
            pass
        try:
            self.client.create_group(self.test_group)
        except:
            pass

    def tearDown(self):
        self.client.delete(self.test_root)
        try:
            self.client.delete_user(self.share2user)
        except:
            pass
        try:
            self.client.delete_group(self.test_group)
        except:
            pass
        self.client.logout()
        shutil.rmtree(self.temp_dir)

    @staticmethod
    def __create_file(target_file, size):
        file_handle = open(target_file, 'w')
        dummy_data = ''
        for i in range(0, 1024):
            dummy_data += 'X'

        for i in range(0, int(size / 1024)):
            # write in 1kb blocks
            file_handle.write(dummy_data)

        dummy_data = ''
        for i in range(0, size % 1024):
            dummy_data += 'X'

        if (len(dummy_data) > 0):
            file_handle.write(dummy_data)

        file_handle.close()

    @data_provider(files_content)
    def test_mkdir(self, file_name, content, subdir):
        """Test subdirectory creation"""
        self.assertTrue(self.client.mkdir(self.test_root + subdir))
        self.assertIsNotNone(self.client.file_info(self.test_root + subdir))

    @data_provider(files_content)
    def test_put_file_contents(self, file_name, content, subdir):
        """Test creating remote file with given contents"""
        self.assertTrue(self.client.mkdir(self.test_root + subdir))
        self.assertTrue(self.client.put_file_contents(self.test_root + subdir + '/' + file_name, content))

    @data_provider(files_content)
    def test_get_file_contents(self, file_name, content, subdir):
        """Test reading remote file"""
        self.assertTrue(self.client.mkdir(self.test_root + subdir))
        self.assertTrue(self.client.put_file_contents(self.test_root + subdir + '/' + file_name, content))
        self.assertEqual(self.client.get_file_contents(self.test_root + subdir + '/' + file_name), content)

    @data_provider(files_content)
    def test_get_file_info(self, file_name, content, subdir):
        """Test getting file info"""
        self.assertTrue(self.client.mkdir(self.test_root + subdir))
        self.assertTrue(self.client.put_file_contents(self.test_root + file_name, content))

        file_info = self.client.file_info(self.test_root + file_name)
        self.assertTrue(isinstance(file_info, nextcloud.FileInfo))
        self.assertEqual(file_info.get_name(), file_name)
        self.assertEqual(file_info.get_path() + '/', self.test_root)
        self.assertEqual(file_info.get_size(), len(content))
        self.assertIsNotNone(file_info.get_etag())
        self.assertEqual(file_info.get_content_type(), 'text/plain')
        self.assertTrue(type(file_info.get_last_modified()) is datetime.datetime)
        self.assertFalse(file_info.is_dir())

        dir_info = self.client.file_info(self.test_root + subdir)
        self.assertTrue(isinstance(dir_info, nextcloud.FileInfo))
        self.assertEqual(dir_info.get_name(), subdir)
        self.assertEqual(file_info.get_path() + '/', self.test_root)
        self.assertIsNone(dir_info.get_size())
        self.assertIsNotNone(dir_info.get_etag())
        self.assertEqual(dir_info.get_content_type(), 'httpd/unix-directory')
        self.assertTrue(type(dir_info.get_last_modified()) is datetime.datetime)
        self.assertTrue(dir_info.is_dir())

    @data_provider(files_content)
    def test_get_file_info_with_properties(self, file_name, content, subdir):
        """Test getting file info with extra properties"""
        self.assertTrue(self.client.mkdir(self.test_root + subdir))
        self.assertTrue(self.client.put_file_contents(self.test_root + file_name, content))

        # ensure we first pass all the properties expected by the tests below, then any extra attributes
        file_info = self.client.file_info(self.test_root + file_name, [
            'd:getlastmodified',
            'd:getcontentlength',
            'd:resourcetype',
            'd:getetag',
            'd:getcontenttype',
            'oc:favorite',
            'oc:owner-id',
            'oc:owner-display-name'
        ])
        self.assertTrue(isinstance(file_info, nextcloud.FileInfo))
        self.assertEqual(file_info.get_name(), file_name)
        self.assertEqual(file_info.get_path() + '/', self.test_root)
        self.assertEqual(file_info.get_size(), len(content))
        self.assertIsNotNone(file_info.get_etag())
        self.assertEqual(file_info.get_content_type(), 'text/plain')
        self.assertTrue(type(file_info.get_last_modified()) is datetime.datetime)
        self.assertFalse(file_info.is_dir())
        self.assertIsNotNone(file_info.attributes['{http://nextcloud.org/ns}favorite'])
        self.assertIsNotNone(file_info.attributes['{http://nextcloud.org/ns}owner-id'])
        self.assertIsNotNone(file_info.attributes['{http://nextcloud.org/ns}owner-display-name'])

        dir_info = self.client.file_info(self.test_root + subdir, [
            'd:getlastmodified',
            'd:getcontentlength',
            'd:resourcetype',
            'd:getetag',
            'd:getcontenttype',
            'oc:favorite',
            'oc:owner-id',
            'oc:owner-display-name'
        ])
        self.assertTrue(isinstance(dir_info, nextcloud.FileInfo))
        self.assertEqual(dir_info.get_name(), subdir)
        self.assertEqual(file_info.get_path() + '/', self.test_root)
        self.assertIsNone(dir_info.get_size())
        self.assertIsNotNone(dir_info.get_etag())
        self.assertEqual(dir_info.get_content_type(), 'httpd/unix-directory')
        self.assertTrue(type(dir_info.get_last_modified()) is datetime.datetime)
        self.assertTrue(dir_info.is_dir())
        self.assertIsNotNone(dir_info.attributes['{http://nextcloud.org/ns}favorite'])
        self.assertIsNotNone(dir_info.attributes['{http://nextcloud.org/ns}owner-id'])
        self.assertIsNotNone(dir_info.attributes['{http://nextcloud.org/ns}owner-display-name'])

    def test_get_file_info_non_existing(self):
        """Test getting file info for non existing file"""
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.client.file_info(self.test_root + 'unexist')
        self.assertEqual(e.exception.status_code, 404)

    def test_get_file_listing(self):
        """Test getting file listing"""
        self.assertTrue(self.client.put_file_contents(self.test_root + 'file one.txt', 'first file'))
        self.assertTrue(self.client.put_file_contents(self.test_root + 'zz+z.txt', 'z file'))
        self.assertTrue(self.client.put_file_contents(self.test_root + u'中文.txt', ''))
        self.assertTrue(self.client.put_file_contents(self.test_root + 'abc.txt', ''))
        self.assertTrue(self.client.mkdir(self.test_root + 'subdir'))
        self.assertTrue(self.client.put_file_contents(self.test_root + 'subdir/in dir.txt', ''))

        listing = self.client.list(self.test_root)
        self.assertEqual(len(listing), 5)
        self.assertEqual(listing[0].get_name(), 'abc.txt')
        self.assertEqual(listing[1].get_name(), 'file one.txt')
        self.assertEqual(listing[2].get_name(), 'subdir')
        self.assertEqual(listing[3].get_name(), 'zz+z.txt')
        self.assertEqual(listing[4].get_name(), u'中文.txt')

        self.assertTrue(listing[2].is_dir())
        self.assertFalse(listing[3].is_dir())

        listing = self.client.list(self.test_root, depth="infinity")
        self.assertEqual(len(listing), 6)
        self.assertEqual(listing[2].get_name(), 'subdir')
        self.assertEqual(listing[3].get_name(), 'in dir.txt')

        listing = self.client.list(self.test_root, depth=2)
        self.assertEqual(len(listing), 6)
        self.assertEqual(listing[2].get_name(), 'subdir')
        self.assertEqual(listing[3].get_name(), 'in dir.txt')

    def test_get_file_listing_with_properties(self):
        """Test getting file listing with extra properties"""
        self.assertTrue(self.client.put_file_contents(self.test_root + 'file one.txt', 'first file'))
        self.assertTrue(self.client.put_file_contents(self.test_root + 'zz+z.txt', 'z file'))
        self.assertTrue(self.client.put_file_contents(self.test_root + u'中文.txt', ''))
        self.assertTrue(self.client.put_file_contents(self.test_root + 'abc.txt', ''))
        self.assertTrue(self.client.mkdir(self.test_root + 'subdir'))
        self.assertTrue(self.client.put_file_contents(self.test_root + 'subdir/in dir.txt', ''))

        listing = self.client.list(self.test_root, depth=1, properties=[
            'd:getlastmodified',
            'd:getcontentlength',
            'd:resourcetype',
            'd:getetag',
            'd:getcontenttype',
            'oc:favorite',
            'oc:owner-id',
            'oc:owner-display-name'
        ])
        self.assertEqual(len(listing), 5)
        self.assertEqual(listing[0].get_name(), 'abc.txt')
        self.assertEqual(listing[1].get_name(), 'file one.txt')
        self.assertEqual(listing[2].get_name(), 'subdir')
        self.assertEqual(listing[3].get_name(), 'zz+z.txt')
        self.assertEqual(listing[4].get_name(), u'中文.txt')

        self.assertTrue(listing[2].is_dir())
        self.assertFalse(listing[3].is_dir())

        for i in range(5):
            self.assertIsNotNone(listing[i].attributes['{http://nextcloud.org/ns}favorite'])
            self.assertIsNotNone(listing[i].attributes['{http://nextcloud.org/ns}owner-id'])
            self.assertIsNotNone(listing[i].attributes['{http://nextcloud.org/ns}owner-display-name'])


    def test_get_file_listing_non_existing(self):
        """Test getting file listing for non existing directory"""
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.client.list(self.test_root + 'unexist')
        self.assertEqual(e.exception.status_code, 404)

    @data_provider(files)
    def test_upload_small_file(self, file_name):
        """Test simple upload"""
        temp_file = self.temp_dir + 'pyoctest.dat'
        self.__create_file(temp_file, 2 * 1024)
        self.assertTrue(self.client.put_file(self.test_root + file_name, temp_file, chunked=False))
        os.unlink(temp_file)

        file_info = self.client.file_info(self.test_root + file_name)
        self.assertIsNotNone(file_info)
        self.assertEqual(file_info.get_size(), 2 * 1024)

    def test_upload_two_chunks(self):
        """Test chunked upload with two chunks"""
        temp_file = self.temp_dir + 'pyoctest.dat'
        self.__create_file(temp_file, 18 * 1024 * 1024 + 1)
        self.assertTrue(self.client.put_file(self.test_root + 'chunk_test.dat', temp_file))
        os.unlink(temp_file)

        file_info = self.client.file_info(self.test_root + 'chunk_test.dat')

        self.assertIsNotNone(file_info)
        self.assertEqual(file_info.get_size(), 18 * 1024 * 1024 + 1)

    def test_upload_chunks_minus_one_byte(self):
        """Test chunked upload minus one byte"""
        temp_file = self.temp_dir + 'pyoctest.dat'
        self.__create_file(temp_file, 2 * 1024 - 1)
        self.assertTrue(self.client.put_file(self.test_root + 'chunk_test.dat', temp_file, chunk_size = 1024))
        os.unlink(temp_file)

        file_info = self.client.file_info(self.test_root + 'chunk_test.dat')

        self.assertIsNotNone(file_info)
        self.assertEqual(file_info.get_size(), 2 * 1024 - 1)

    @data_provider(files)
    def test_upload_big_file(self, file_name):
        """Test chunked upload"""
        temp_file = self.temp_dir + 'pyoctest.dat'
        self.__create_file(temp_file, 22 * 1024 * 1024)
        self.assertTrue(self.client.put_file(self.test_root + file_name, temp_file))
        os.unlink(temp_file)

        file_info = self.client.file_info(self.test_root + file_name)
        self.assertIsNotNone(file_info)
        self.assertEqual(file_info.get_size(), 22 * 1024 * 1024)

    def test_upload_timestamp(self):
        # TODO: test with keeping timestamp and not keeping it
        pass

    def test_upload_directory(self):
        temp_dir = self.temp_dir + 'pyoctest.dir/'
        os.mkdir(temp_dir)
        os.mkdir(temp_dir + 'levelone')
        os.mkdir(temp_dir + 'levelone/leveltwo')

        self.__create_file(temp_dir + 'file1.dat', 2 * 1024)
        self.__create_file(temp_dir + 'file2.dat', 22 * 1024 * 1024)
        self.__create_file(temp_dir + 'levelone/file3.dat', 22 * 1024 * 1024)
        self.__create_file(temp_dir + 'levelone/leveltwo/file4.dat', 8 * 1024 * 1024)
        self.__create_file(temp_dir + u'levelone/文件.dat', 7 * 1024 * 1024)

        self.assertTrue(self.client.mkdir(self.test_root + 'subdir'))
        self.assertTrue(self.client.put_directory(self.test_root + 'subdir', temp_dir))

        self.assertIsNotNone(self.client.file_info(self.test_root + 'subdir/pyoctest.dir'))
        self.assertIsNotNone(self.client.file_info(self.test_root + 'subdir/pyoctest.dir/file1.dat'))
        self.assertIsNotNone(self.client.file_info(self.test_root + 'subdir/pyoctest.dir/file2.dat'))
        self.assertIsNotNone(self.client.file_info(self.test_root + 'subdir/pyoctest.dir/levelone/file3.dat'))
        self.assertIsNotNone(self.client.file_info(self.test_root + 'subdir/pyoctest.dir/levelone/leveltwo/file4.dat'))
        self.assertIsNotNone(self.client.file_info(self.test_root + u'subdir/pyoctest.dir/levelone/文件.dat'))

    @data_provider(files_content)
    def test_download_file(self, file_name, content, subdir):
        """Test file download"""
        temp_file = self.temp_dir + 'pyoctest.dat'
        self.assertTrue(self.client.mkdir(self.test_root + subdir))
        self.assertTrue(self.client.put_file_contents(self.test_root + subdir + '/' + file_name, content))

        self.assertTrue(self.client.get_file(self.test_root + subdir + '/' + file_name, temp_file))

        f = open(temp_file, 'rb')
        s = f.read()
        f.close()
        os.unlink(temp_file)
        self.assertEqual(s, content)

    def test_download_dir(self):
        import zipfile
        """Test directory download as zip"""
        self.assertTrue(self.client.mkdir(self.test_root + 'subdir'))
        self.assertTrue(self.client.put_file_contents(self.test_root + 'subdir/test.txt', 'hello world!'))
        # Note: this can only work properly with OC 7
        #self.assertTrue(self.client.put_file_contents(self.test_root + 'subdir/文件.txt', '你好世界!'))

        temp_file = self.temp_dir + 'pyoctest.zip'
        self.assertTrue(self.client.get_directory_as_zip(self.test_root, temp_file))

        self.assertTrue(os.stat(temp_file))

        zip_info = zipfile.ZipFile(temp_file)
        listing = zip_info.namelist()

        self.assertEqual(len(listing), 3)
        os.unlink(temp_file)

    @data_provider(files_content)
    def test_delete_file(self, file_name, content, subdir):
        """Test file deletion"""
        self.assertTrue(self.client.mkdir(self.test_root + subdir))
        self.assertTrue(self.client.put_file_contents(self.test_root + subdir + '/' + file_name, content))
        self.assertTrue(self.client.delete(self.test_root + subdir + '/' + file_name))
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.client.file_info(self.test_root + subdir + '/' + file_name)
        self.assertEqual(e.exception.status_code, 404)

    @data_provider(files_content)
    def test_delete_dir(self, file_name, content, subdir):
        """Test directory deletion"""
        self.assertTrue(self.client.mkdir(self.test_root + subdir))
        self.assertTrue(self.client.put_file_contents(self.test_root + subdir + '/' + file_name, content))
        self.assertTrue(self.client.delete(self.test_root + subdir))
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.client.file_info(self.test_root + subdir + '/' + file_name)
        self.assertEqual(e.exception.status_code, 404)
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.client.file_info(self.test_root + subdir)
        self.assertEqual(e.exception.status_code, 404)

    def test_move_rename_in_place(self):
        """Test rename in place"""

        self.assertTrue(
            self.client.put_file_contents(
                self.test_root + 'rename this file!.txt',
                'to rename'
            )
        )
        self.assertTrue(
            self.client.move(
                self.test_root + 'rename this file!.txt',
                self.test_root + 'renamed in place.txt'
            )
        )
        self.assertEqual(
            self.client.get_file_contents(
                self.test_root + 'renamed in place.txt'
            ),
            b'to rename'
        )

    def test_move_and_rename(self):
        """Test rename into subdir"""

        self.assertTrue(
            self.client.put_file_contents(
                self.test_root + 'move and rename.txt',
                'first file'
            )
        )
        self.assertTrue(
            self.client.mkdir(
                self.test_root + 'subdir'
            )
        )
        self.assertTrue(
            self.client.move(
                self.test_root + 'move and rename.txt',
                self.test_root + 'subdir/file renamed.txt'
            )
        )
        self.assertEqual(
            self.client.get_file_contents(
                self.test_root + 'subdir/file renamed.txt'
            ),
            b'first file'
        )

    def test_move_to_dir(self):
        """Test move into directory"""

        self.assertTrue(
            self.client.put_file_contents(
                self.test_root + 'movetodir.txt',
                'z file'
            )
        )
        # move to subdir
        self.assertTrue(
            self.client.mkdir(
                self.test_root + 'subdir'
            )
        )
        self.assertTrue(
            self.client.move(
                self.test_root + 'movetodir.txt',
                self.test_root + 'subdir/'
            )
        )
        self.assertEqual(
            self.client.get_file_contents(
                self.test_root + 'subdir/movetodir.txt'
            ),
            b'z file'
        )

    def test_move_subdir(self):
        """Test move subdir"""

        # subdir to move
        self.assertTrue(
            self.client.mkdir(
                self.test_root + 'subdir_to_move'
            )
        )
        self.assertTrue(
            self.client.put_file_contents(
                self.test_root + 'subdir_to_move/file two.txt',
                'second file'
            )
        )
        self.assertTrue(
            self.client.mkdir(
                self.test_root + 'subdir'
            )
        )
        self.assertTrue(
            self.client.move(
                self.test_root + 'subdir_to_move',
                self.test_root + 'subdir/'
            )
        )

    def test_rename_unicode(self):
        """Test rename unicode"""

        # rename
        self.assertTrue(
            self.client.put_file_contents(
                self.test_root + u'中文.txt',
                '1'
            )
        )
        self.assertTrue(
            self.client.move(
                self.test_root + u'中文.txt',
                self.test_root + u'更多中文.txt'
            )
        )
        self.assertEqual(
            self.client.get_file_contents(
                self.test_root + u'更多中文.txt'
            ),
            b'1'
        )

    def test_move_unicode(self):
        """Test move unicode to dir"""
        self.assertTrue(
            self.client.put_file_contents(
                self.test_root + u'中文.txt',
                '2'
            )
        )
        self.assertTrue(
            self.client.mkdir(
                self.test_root + 'subdir'
            )
        )
        self.assertTrue(
            self.client.move(
                self.test_root + u'中文.txt',
                self.test_root + u'subdir/中文.txt'
            )
        )
        self.assertEqual(
            self.client.get_file_contents(
                self.test_root + u'subdir/中文.txt'
            ),
            b'2'
        )

    def test_move_to_non_existing_dir(self):
        """Test error when moving to non existing dir"""

        self.assertTrue(
            self.client.put_file_contents(
                self.test_root + 'move not possible.txt',
                'x'
            )
        )
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.client.move(
                self.test_root + 'move not possible.txt',
                self.test_root + 'non-existing-dir/subdir/x.txt'
            )
        self.assertEqual(e.exception.status_code, 409)

        self.assertEqual(
            self.client.get_file_contents(
                self.test_root + 'move not possible.txt'
            ),
            b'x'
        )

    def test_copy_in_place(self):
        """Test copy in place"""

        self.assertTrue(
            self.client.put_file_contents(
                self.test_root + 'copy this file!.txt',
                'to copy'
            )
        )
        self.assertTrue(
            self.client.copy(
                self.test_root + 'copy this file!.txt',
                self.test_root + 'copied in place.txt'
            )
        )
        self.assertEqual(
            self.client.get_file_contents(
                self.test_root + 'copied in place.txt'
            ),
            b'to copy'
        )

        self.assertEqual(
            self.client.get_file_contents(
                self.test_root + 'copy this file!.txt'
            ),
            b'to copy'
        )

    def test_copy_into_subdir(self):
        """Test copy into subdir"""

        self.assertTrue(
            self.client.put_file_contents(
                self.test_root + 'copy into subdir.txt',
                'first file'
            )
        )
        self.assertTrue(
            self.client.mkdir(
                self.test_root + 'subdir'
            )
        )
        self.assertTrue(
            self.client.copy(
                self.test_root + 'copy into subdir.txt',
                self.test_root + 'subdir/file copied.txt'
            )
        )
        self.assertEqual(
            self.client.get_file_contents(
                self.test_root + 'subdir/file copied.txt'
            ),
            b'first file'
        )
        self.assertEqual(
            self.client.get_file_contents(
                self.test_root + 'copy into subdir.txt'
            ),
            b'first file'
        )

    def test_copy_unicode(self):
        """Test copy unicode to dir"""
        self.assertTrue(
            self.client.put_file_contents(
                self.test_root + u'दिलै होस छोरा होस.txt',
                'content'
            )
        )
        self.assertTrue(
            self.client.mkdir(
                self.test_root + 'subdir'
            )
        )
        self.assertTrue(
            self.client.copy(
                self.test_root + u'दिलै होस छोरा होस.txt',
                self.test_root + u'subdir/दिलै होस छोरा होस.txt'
            )
        )
        self.assertEqual(
            self.client.get_file_contents(
                self.test_root + u'subdir/दिलै होस छोरा होस.txt'
            ),
            b'content'
        )
        self.assertEqual(
            self.client.get_file_contents(
                self.test_root + u'दिलै होस छोरा होस.txt'
            ),
            b'content'
        )

    def test_copy_to_non_existing_dir(self):
        """Test error when copy to non existing dir"""

        self.assertTrue(
            self.client.put_file_contents(
                self.test_root + 'copy not possible.txt',
                'x'
            )
        )
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.client.copy(
                self.test_root + 'copy not possible.txt',
                self.test_root + 'non-existing-dir/subdir/x.txt'
            )
        self.assertEqual(e.exception.status_code, 409)

        self.assertEqual(
            self.client.get_file_contents(
                self.test_root + 'copy not possible.txt'
            ),
            b'x'
        )

    @data_provider(files)
    def test_share_with_link(self, file_name):
        """Test sharing a file with link"""

        path = self.test_root + file_name
        self.assertTrue(self.client.put_file_contents(path, 'hello world!'))

        share_info = self.client.share_file_with_link(path, public_upload=False, password='AnvvsP1234', name='Test Link')

        self.assertTrue(self.client.is_shared(path))
        self.assertTrue(isinstance(share_info, nextcloud.ShareInfo))
        self.assertTrue(type(share_info.get_id()) is int)
        self.assertEqual(share_info.get_path(), path)
        self.assertEqual(share_info.get_name(), 'Test Link')
        self.assertTrue(type(share_info.get_link()) is str)
        self.assertTrue(type(share_info.get_token()) is str)

    def test_share_with_link_non_existing_file(self):
        """Test sharing a file with link"""
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.client.share_file_with_link(self.test_root + 'unexist.txt')
        self.assertEqual(e.exception.status_code, 404)

    @data_provider(files)
    def test_share_with_user(self, file_name):
        """Test sharing a file to user"""

        path = self.test_root + file_name
        self.assertTrue(self.client.put_file_contents(path, 'hello world!'))

        share_info = self.client.share_file_with_user(path, self.share2user)

        self.assertTrue(self.client.is_shared(path))
        self.assertTrue(isinstance(share_info, nextcloud.ShareInfo))
        self.assertEqual(share_info.get_path(), path)
        self.assertTrue(type(share_info.get_id()) is int)
        self.assertEqual(share_info.get_permissions(), 1)

        shareclient = nextcloud.Client(Config['owncloud_url'])
        shareclient.login(self.share2user, self.share2userPwd)
        share2_info = shareclient.get_shares(
            "/", shared_with_me=True)[0].share_info
        self.assertEqual(share2_info["uid_owner"], Config['owncloud_login'])
        self.assertEqual(file_name, share2_info["file_target"][1:])
        shareclient.logout()

        self.assertTrue(self.client.delete(path))

    @data_provider(files)
    def test_share_with_group(self, file_name):
        """Test sharing a file to a group"""

        path = self.test_root + file_name
        self.assertTrue(self.client.put_file_contents(path, 'hello world!'))

        share_info = self.client.share_file_with_group(path, self.test_group, perms=31)

        self.assertTrue(self.client.is_shared(path))
        self.assertTrue(isinstance(share_info, nextcloud.ShareInfo))
        self.assertEqual(share_info.get_path(), path)
        self.assertTrue(type(share_info.get_id()) is int)
        self.assertEqual(share_info.get_permissions(), 31)
        self.assertTrue(self.client.delete(path))

    @data_provider(files)
    def test_delete_share(self, file_name):
        """Test deleting a share (by id)"""

        path = self.test_root + file_name
        self.assertTrue(self.client.put_file_contents(path, 'hello world!'))

        share_info = self.client.share_file_with_user(path, self.share2user)

        self.assertTrue(self.client.is_shared(path))
        self.assertIsNotNone(self.client.delete_share(share_info.get_id()))
        self.assertFalse(self.client.is_shared(path))

    def test_is_shared_non_existing_path(self):
        """Test is_shared - path does not exist"""
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.client.is_shared(self.test_root + 'does_not_exist')
        self.assertEqual(e.exception.status_code, 404)

    def test_is_shared_not_shared_path(self):
        """Test is_shared - path does exist, but it's not shared yet"""
        path = self.test_root + 'not_shared_path.txt'
        self.assertTrue(self.client.put_file_contents(path, 'hello world!'))
        self.assertFalse(self.client.is_shared(path))

    @data_provider(files)
    def test_is_shared(self, file_name):
        """Test is_shared"""
        path = self.test_root + file_name
        self.assertTrue(self.client.put_file_contents(path, 'hello world!'))

        self.client.share_file_with_link(path)
        self.assertTrue(self.client.is_shared(path))
        self.assertTrue(self.client.delete(path))

    @data_provider(files)
    def test_get_share_user(self, file_name):
        """Test get_share() for user share"""
        path = self.test_root + file_name
        self.assertTrue(self.client.put_file_contents(path, 'hello world!'))

        sinfo_run = self.client.share_file_with_user(
            path,
            self.share2user,
            perms=self.client.OCS_PERMISSION_READ | self.client.OCS_PERMISSION_SHARE
        )
        sinfo = self.client.get_share(sinfo_run.get_id())
        self.assertIsInstance(sinfo, nextcloud.ShareInfo)
        share_id = sinfo.get_id()
        self.assertGreater(share_id, 0)
        self.assertEqual(sinfo_run.get_id(), share_id)
        self.assertIsInstance(sinfo.get_id(), int)
        self.assertEqual(sinfo.get_share_type(), self.client.OCS_SHARE_TYPE_USER)
        self.assertEqual(sinfo.get_share_with(), self.share2user)
        self.assertEqual(sinfo.get_path(), path)
        self.assertEqual(
            sinfo.get_permissions(),
            self.client.OCS_PERMISSION_READ | self.client.OCS_PERMISSION_SHARE
        )
        self.assertIsInstance(sinfo.get_share_time(), datetime.datetime)
        self.assertIsNone(sinfo.get_expiration())
        self.assertIsNone(sinfo.get_token())
        self.assertEqual(sinfo.get_uid_owner(), Config['owncloud_login'])
        self.assertIsInstance(sinfo.get_displayname_owner(), six.string_types)

    @data_provider(files)
    def test_get_share_public_link(self, file_name):
        """Test get_share() for public link share"""
        path = self.test_root + file_name
        self.assertTrue(self.client.put_file_contents(path, 'hello world!'))

        sinfo_run = self.client.share_file_with_link(path)
        sinfo = self.client.get_share(sinfo_run.get_id())
        self.assertIsInstance(sinfo, nextcloud.ShareInfo)
        self.assertIsNotNone(sinfo)
        share_id = sinfo.get_id()
        self.assertGreater(share_id, 0)
        self.assertEqual(sinfo_run.get_id(), share_id)
        self.assertIsInstance(sinfo.get_id(), int)
        self.assertEqual(sinfo.get_share_type(), self.client.OCS_SHARE_TYPE_LINK)
        self.assertIsNone(sinfo.get_share_with())
        self.assertEqual(sinfo.get_path(), path)
        self.assertEqual(sinfo.get_permissions(), self.client.OCS_PERMISSION_READ)
        self.assertIsInstance(sinfo.get_share_time(), datetime.datetime)
        self.assertIsNone(sinfo.get_expiration())
        self.assertIsInstance(sinfo.get_token(), six.string_types)
        self.assertEqual(sinfo.get_uid_owner(), Config['owncloud_login'])
        self.assertIsInstance(sinfo.get_displayname_owner(), six.string_types)

    def test_get_share_non_existing(self):
        """Test get_share - share with specified id does not exist"""
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.client.get_share(-1)
        self.assertEqual(e.exception.status_code, 404)

    def test_get_shares_non_existing_path(self):
        """Test get_shares - path does not exist"""
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.client.get_shares(self.test_root + 'does_not_exist')
        self.assertEqual(e.exception.status_code, 404)

    @data_provider(files)
    def test_get_shares(self, file_name):
        """Test get_shares"""
        self.assertTrue(self.client.put_file_contents(self.test_root + file_name, 'hello world!'))

        self.client.share_file_with_link(self.test_root + file_name)
        shares = self.client.get_shares(self.test_root + file_name)
        self.assertIsNotNone(shares)
        self.assertIsInstance(shares, list)

        shares = None
        with self.assertRaises(nextcloud.ResponseError) as e:
            shares = self.client.get_shares(self.test_root + file_name, subfiles=True)
        self.assertIsNone(shares)
        self.assertEqual(e.exception.status_code, 400)

        shares = self.client.get_shares(self.test_root, reshares=True, subfiles=True)
        self.assertIsNotNone(shares)
        self.assertIsInstance(shares, list)
        self.assertGreater(len(shares), 0)

        self.assertTrue(self.client.put_file_contents(self.test_root + file_name + '2.txt', 'hello world!'))
        self.client.share_file_with_link(self.test_root + file_name + '2.txt')
        shares = self.client.get_shares(self.test_root, reshares=True, subfiles=True)
        self.assertIsNotNone(shares)
        self.assertIsInstance(shares, list)
        self.assertGreater(len(shares), 1)

    def test_get_shares_empty(self):
        """Test get shares with empty result"""
        file_name = 'test.txt'
        self.assertTrue(self.client.put_file_contents(self.test_root + file_name, 'hello world!'))

        # Get all shares
        shares = self.client.get_shares()
        self.assertEqual(shares, [])

    def test_update_share_wo_params(self):
        self.assertFalse(self.client.update_share(0))

    def test_update_share_user(self):
        """Test updating a share parameters - user share"""
        path = self.test_root + 'update_share_user.txt'
        self.assertTrue(self.client.put_file_contents(path, 'hello world!'))

        share_info = self.client.share_file_with_user(path, self.share2user)
        share_id = share_info.get_id()
        maxPerms = self.client.OCS_PERMISSION_READ + self.client.OCS_PERMISSION_UPDATE + self.client.OCS_PERMISSION_SHARE
        self.assertTrue(self.client.update_share(share_id, perms=maxPerms))
        perms = self.client.get_shares(path)[0].get_permissions()
        # now the permissions should be OCS_PERMISSION_ALL,
        # because we've shared it with a user
        self.assertEqual(int(perms), maxPerms)
        self.assertTrue(self.client.delete_share(share_id))

    def test_update_share_public(self):
        """Test updating a share parameters - public share"""
        path = self.test_root + 'update_share_public.txt'
        self.assertTrue(self.client.put_file_contents(path, 'hello world!'))

        share_info = self.client.share_file_with_link(path)
        share_id = share_info.get_id()
        self.assertTrue(self.client.update_share(share_id, perms=self.client.OCS_PERMISSION_ALL))
        perms = self.client.get_shares(path)[0].get_permissions()
        # permissions should still be OCS_PERMISSION_READ not OCS_PERMISSION_ALL,
        # because it's a public share
        self.assertEqual(int(perms), self.client.OCS_PERMISSION_READ)
        self.assertTrue(self.client.delete_share(share_id))

    def test_update_share_public_upload(self):
        """Test updating a share parameters - public upload"""
        path = self.test_root + 'update_share_public_upload'
        self.client.mkdir(path)

        share_info = self.client.share_file_with_link(path)
        share_id = share_info.get_id()
        self.assertTrue(self.client.update_share(share_id, public_upload=True))
        perms = self.client.get_shares(path)[0].get_permissions()

        # make sure the server did receive the upload permission
        self.assertTrue(perms & self.client.OCS_PERMISSION_CREATE > 0)
        self.assertTrue(perms & self.client.OCS_PERMISSION_UPDATE > 0)

        # test reverting to read only
        self.assertTrue(self.client.update_share(share_id, public_upload=False))
        perms = self.client.get_shares(path)[0].get_permissions()
        self.assertEqual(int(perms), self.client.OCS_PERMISSION_READ)
        self.assertTrue(self.client.delete_share(share_id))

    def test_update_share_password(self):
        """Test updating a share parameters - password"""
        path = self.test_root + 'update_share_password'
        self.client.mkdir(path)

        share_info = self.client.share_file_with_link(path)
        share_id = share_info.get_id()
        self.assertTrue(self.client.update_share(share_id, password="2hard2guess"))
        share_info = self.client.get_shares(path)[0]
        self.assertTrue(type(share_info.get_share_with_displayname()) is str)
        self.assertTrue(self.client.delete_share(share_id))

@skipIfDavVersionLessThan(1.0)
class TestFileAccessDav1(TestFileAccess):
    def get_dav_endpoint_version(self):
        return 1

class TestPrivateDataAccess(unittest.TestCase):

    def attrs():
        return (
            ('attr1', 'value1'),
            ('attr+plus space', 'value+plus space and/slash'),
            (u'属性1', u'值对1')
        )

    def setUp(self):
        self.client = nextcloud.Client(Config['owncloud_url'])
        self.client.login(Config['owncloud_login'], Config['owncloud_password'])
        self.app_name = Config['app_name']

    def tearDown(self):
        # TODO: delete all attributes ?
        self.client.logout()

    @data_provider(attrs)
    def test_set_attribute(self, attr1, value1):
        """Test setting an attribute"""
        self.assertTrue(self.client.set_attribute(self.app_name, attr1, value1))
        self.assertTrue(self.client.delete_attribute(self.app_name, attr1))

    @data_provider(attrs)
    def test_get_attribute(self, attr1, value1):
        """Test getting an attribute"""
        self.assertTrue(self.client.set_attribute(self.app_name, attr1, value1))

        self.assertEqual(self.client.get_attribute(self.app_name, attr1), value1)
        self.assertEqual(self.client.get_attribute(self.app_name), [(attr1, value1)])
        self.assertTrue(self.client.delete_attribute(self.app_name, attr1))

    def test_get_non_existing_attribute(self):
        """Test getting the value of a non existing attribute"""
        self.assertIsNone(self.client.get_attribute(self.app_name, 'unexist'))

    @data_provider(attrs)
    def test_set_attribute_empty(self, attr1, value1):
        """Test setting an attribute to an empty value"""
        self.assertTrue(self.client.set_attribute(self.app_name, attr1, ''))
        self.assertEqual(self.client.get_attribute(self.app_name, attr1), '')
        self.assertEqual(self.client.get_attribute(self.app_name), [(attr1, '')])
        self.assertTrue(self.client.delete_attribute(self.app_name, attr1))

    @data_provider(attrs)
    def test_delete_attribute(self, attr1, value1):
        """Test deleting an attribute"""
        self.assertTrue(self.client.set_attribute(self.app_name, attr1, value1))
        self.assertEqual(self.client.get_attribute(self.app_name, attr1), value1)

        self.assertTrue(self.client.delete_attribute(self.app_name, attr1))

        self.assertIsNone(self.client.get_attribute(self.app_name, attr1))
        self.assertEqual(self.client.get_attribute(self.app_name), [])


class TestUserAndGroupActions(unittest.TestCase):

    def setUp(self):
        self.client = nextcloud.Client(Config['owncloud_url'])
        self.client.login(Config['owncloud_login'], Config['owncloud_password'])
        self.groups_to_create = Config['groups_to_create']
        self.not_existing_group = Config['not_existing_group']
        self.test_group = Config['test_group']
        self.share2user = Config['owncloud_share2user']
        self.share2userPwd = 'Avcpwd4l!'
        try:
            self.apps = self.client.get_apps()
            if not self.apps['provisioning_api']:
                raise unittest.SkipTest("no API")
        except nextcloud.ResponseError:
            raise unittest.SkipTest("no API")

        try:
            self.client.create_user(self.share2user, self.share2userPwd)
        except:
            pass
        try:
            self.client.create_group(self.test_group)
        except:
            pass

    def tearDown(self):
        for group in self.groups_to_create:
            self.assertTrue(self.client.delete_group(group))

        self.assertTrue(self.client.remove_user_from_group(self.share2user,self.test_group))
        try:
            self.client.delete_user(self.share2user)
        except:
            pass
        try:
            self.client.delete_group(self.test_group)
        except:
            pass

        self.client.logout()

    def test_user_exists(self):
        self.assertTrue(self.client.user_exists(Config['owncloud_login']))
        try:
            self.client.create_user('ghost_user', 'ghost_pass')
            self.client.delete_user('ghost_user')
            self.assertFalse(self.client.user_exists('ghost_user'))
        except:
            pass

    def test_get_user(self):
        output = self.client.get_user(Config['owncloud_login'])
        expected_output =   {'displayname': 'admin',
                             'enabled': 'true',
                             'email': None,
                             'quota': {'total': '309355267452',
                                       'relative': '0',
                                       'used': '3261820',
                                       'free': '309352005632'}
                            }
        self.assertEqual(output['displayname'], expected_output['displayname'])
        self.assertEqual(output['enabled'], expected_output['enabled'])
        self.assertEqual(output['email'], expected_output['email'])
        self.assertTrue('total' in output['quota'])
        self.assertTrue('relative' in output['quota'])
        self.assertTrue('used' in output['quota'])
        self.assertTrue('free' in output['quota'])

    def test_search_users(self):
        user_name = Config['owncloud_login']
        users = self.client.search_users(user_name[:-1])
        self.assertIn(user_name, users)

    def test_get_users(self):
        user_name = Config['owncloud_login']
        users = self.client.get_users()
        self.assertIn(user_name, users)

    def test_set_user_attribute(self):
        try:
            self.client.create_user('ghost_user', 'ghost_pass')
        except:
            self.client.delete_user('ghost_user')
            self.client.create_user('ghost_user', 'ghost_pass')
        self.assertTrue(self.client.set_user_attribute('ghost_user','email','test@inf.org'))
        self.assertTrue(self.client.set_user_attribute('ghost_user','password','secret7363*'))
        self.assertEqual(self.client.get_user('ghost_user')['email'], 'test@inf.org')
        self.client.delete_user('ghost_user')

        with self.assertRaises(nextcloud.OCSResponseError) as e:
            self.client.set_user_attribute(self.share2user,'email',"äöüää_sfsdf+$%/)%&=")
        self.assertEqual(e.exception.status_code, 102)
        #try to catch with general ResponseError
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.client.set_user_attribute(self.share2user,'email',"äöüää_sfsdf+$%/)%&=")
        self.assertEqual(e.exception.status_code, 102)

    def test_create_existing_user(self):
        with self.assertRaises(nextcloud.OCSResponseError) as e:
            self.client.create_user(self.share2user, self.share2userPwd)
        self.assertEqual(e.exception.status_code, 102)
        # try to catch with general ResponseError
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.client.create_user(self.share2user, self.share2userPwd)
        self.assertEqual(e.exception.status_code, 102)

    def test_create_groups(self):
        for group in self.groups_to_create:
            self.assertTrue(self.client.create_group(group))
            self.assertTrue(self.client.group_exists(group))
            # try to create them again, that should raise and OCSResponseError with code 102
            with self.assertRaises(nextcloud.OCSResponseError) as e:
                self.client.create_group(group)
            self.assertEqual(e.exception.status_code, 102)
            #try to catch with general ResponseError
            with self.assertRaises(nextcloud.ResponseError) as e:
                self.client.create_group(group)
            self.assertEqual(e.exception.status_code, 102)

    def test_get_groups(self):
        test_group = Config['test_group']
        not_existing_group = Config['not_existing_group']
        groups = self.client.get_groups()
        self.assertIn(test_group, groups)
        self.assertNotIn(not_existing_group, groups)

    def test_get_group_members(self):
        user_name = Config['owncloud_login']
        test_group = Config['test_group']
        admin_members = self.client.get_group_members('admin')
        test_members = self.client.get_group_members(test_group)
        self.assertIn(user_name, admin_members)
        self.assertNotIn(user_name, test_members)

    def test_not_existing_group(self):
        self.assertFalse(self.client.group_exists(self.not_existing_group))

    def test_add_user_to_group_remove_user_from_group(self):
        self.assertFalse(self.client.user_is_in_group(self.share2user,self.test_group))
        self.assertTrue(self.client.add_user_to_group(self.share2user,self.test_group))
        self.assertTrue(self.client.user_is_in_group(self.share2user,self.test_group))

        # try to add the user to a not existing group, that should raise and OCSResponseError with code 102
        with self.assertRaises(nextcloud.OCSResponseError) as e:
            self.client.add_user_to_group(self.share2user,self.not_existing_group)
        self.assertEqual(e.exception.status_code, 102)
        # try to catch with general ResponseError
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.client.add_user_to_group(self.share2user,self.not_existing_group)
        self.assertEqual(e.exception.status_code, 102)

        self.assertTrue(self.client.remove_user_from_group(self.share2user,self.test_group))
        self.assertFalse(self.client.user_is_in_group(self.share2user,self.test_group))

        # try to remove the user from a not existing group, that should raise and OCSResponseError with code 102
        with self.assertRaises(nextcloud.OCSResponseError) as e:
            self.client.remove_user_from_group(self.share2user,self.not_existing_group)
        self.assertEqual(e.exception.status_code, 102)
        # try to catch with general ResponseError
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.client.remove_user_from_group(self.share2user,self.not_existing_group)
        self.assertEqual(e.exception.status_code, 102)

        # try to remove user without giving group name
        with self.assertRaises(nextcloud.OCSResponseError) as e:
            self.client.remove_user_from_group(self.share2user,'')
        self.assertEqual(e.exception.status_code, 101)

        # try to remove not existing user from a group
        with self.assertRaises(nextcloud.OCSResponseError) as e:
            self.client.remove_user_from_group("iGuessThisUserNameDoesNotExistInTheSystem",self.test_group)
        self.assertEqual(e.exception.status_code, 103)


class TestApps(unittest.TestCase):

    def setUp(self):
        self.client = nextcloud.Client(Config['owncloud_url'])
        self.client.login(Config['owncloud_login'], Config['owncloud_password'])

    def test_enable_app_disable_app(self):
        self.assertTrue(self.client.enable_app('activity'))
        self.assertTrue(self.client.disable_app('activity'))

    def tearDown(self):
        self.client.logout()


class TestGetConfig(unittest.TestCase):

    def setUp(self):
        self.client = nextcloud.Client(Config['owncloud_url'])
        self.client.login(Config['owncloud_login'], Config['owncloud_password'])

    def test_get_config(self):
        """Test get_config() function"""
        self.assertIsNotNone(self.client.get_config())

    def test_get_version(self):
        """Test get_version() function"""
        version = self.client.get_version()
        self.assertIsNotNone(version)
        version_parts = version.split('.')
        self.assertGreaterEqual(int(version_parts[0]), 5)

    def test_get_capabilities(self):
        """Test get_capabilities() function"""
        caps = self.client.get_capabilities()
        # files app is always enabled
        self.assertIsNotNone(caps['files'])
        # and always has big file chunking enabled
        self.assertEqual(caps['files']['bigfilechunking'], '1')

    def tearDown(self):
        self.client.logout()


class TestLogin(unittest.TestCase):

    def setUp(self):
        self.client = nextcloud.Client(Config['owncloud_url'])

    def test_login(self):
        with self.assertRaises(nextcloud.HTTPResponseError) as e:
            self.client.login("iGuessThisUserNameDoesNotExistInTheSystem","iGuessThisUserNameDoesNotExistInTheSystem")
        self.assertEqual(e.exception.status_code, 401)
        self.client.login(Config['owncloud_login'], Config['owncloud_password'])

    def tearDown(self):
        self.client.logout()


class TestOCSRequest(unittest.TestCase):

    def setUp(self):
        self.client = nextcloud.Client(Config['owncloud_url'])
        self.client.login(Config['owncloud_login'], Config['owncloud_password'])

    def test_make_request(self):
        kwargs = {
            'accepted_codes': [100]
        }
        self.client.make_ocs_request(
            'GET',
            '',
            'config',
            **kwargs
        )

    def test_make_request_fail_unaccepted_code(self):
        kwargs = {
            'accepted_codes': [102]
        }
        with self.assertRaises(nextcloud.OCSResponseError) as e:
            self.client.make_ocs_request(
                'GET',
                '',
                'config',
                **kwargs
            )
        self.assertEqual(e.exception.status_code, 100)

    def tearDown(self):
        self.client.logout()

class TestPublicFolder(unittest.TestCase):

    def get_dav_endpoint_version(self):
        return 0

    def files():
        return (
            ['test.txt'],
            ['test space and + and #.txt'],
            [u'文件.txt']
        )

    def files_content():
        return (
            ['test.txt', b'Hello world!', 'subdir'],
            ['test space and + and #.txt', b'Hello space with+plus#hash!', 'subdir with space + plus and #hash'],
            [u'文件.txt', u'你好世界'.encode('utf-8'), u'文件夹']
        )

    
    def setUp(self):
        self.temp_dir = tempfile.gettempdir() + '/pyocclient_test%s-%s/' % (int(time.time()), random.randint(1, 1000))
        os.mkdir(self.temp_dir)

        self.client = nextcloud.Client(Config['owncloud_url'], dav_endpoint_version=self.get_dav_endpoint_version())
        self.client.login(Config['owncloud_login'], Config['owncloud_password'])
        self.test_root = Config['test_root']
        if not self.test_root[-1] == '/':
            self.test_root += '/'
        if not self.test_root[0] == '/':
            self.test_root = '/' + self.test_root

        # setting up the root dir
        self.client.mkdir(self.test_root)
        self.download_testfile = 'download_pyoctest.dat'
        self.upload_testfile = 'upload_pyoctest.dat'
        self.folder_password = 'secret'
        temp_file = self.temp_dir + self.download_testfile

        # create pathing for different scenarios
        self.public_folder_download = self.test_root + 'public_share_download/' # file download only
        self.public_folder_upload = self.test_root + 'public_share_upload/' # file upload only
        self.pw_public_folder_download = self.test_root + 'pw_public_share_download/' # password protected download
        self.pw_public_folder_upload = self.test_root + 'pw_public_share_upload/' # password protected upload

        # make folder with public download only
        self.client.mkdir(self.public_folder_download)
        public_share_info = self.client.share_file_with_link(self.public_folder_download, perms=self.client.OCS_PERMISSION_READ)
        self.public_download_link = public_share_info.get_link()
        self.public_download_token = public_share_info.get_token()

        # populate folder with data
        temp_file = self.temp_dir + self.download_testfile
        self.__create_file(temp_file, 2 * 1024)
        self.client.put_file(self.public_folder_download + self.download_testfile, temp_file, chunked=False)

        # make folder with public upload only
        self.client.mkdir(self.public_folder_upload)
        public_share_info = self.client.share_file_with_link(self.public_folder_upload, perms=self.client.OCS_PERMISSION_CREATE)
        self.public_upload_link = public_share_info.get_link()
        self.public_upload_token = public_share_info.get_token()

        # make password protected folder with public download only
        self.client.mkdir(self.pw_public_folder_download)
        public_share_info = self.client.share_file_with_link(self.pw_public_folder_download, perms=self.client.OCS_PERMISSION_READ, password=self.folder_password)
        self.pw_public_download_link = public_share_info.get_link()
        self.pw_public_download_token = public_share_info.get_token()

        # populate password protected folder with data
        temp_file = self.temp_dir + self.download_testfile
        self.__create_file(temp_file, 2 * 1024)
        self.client.put_file(self.pw_public_folder_download + self.download_testfile, temp_file, chunked=False)

        # make password protected folder with public upload only
        self.client.mkdir(self.pw_public_folder_upload)
        public_share_info = self.client.share_file_with_link(self.pw_public_folder_upload, perms=self.client.OCS_PERMISSION_CREATE, password=self.folder_password)
        self.pw_public_upload_link = public_share_info.get_link()
        self.pw_public_upload_token = public_share_info.get_token()

    def tearDown(self):
        self.client.delete(self.test_root)
        self.client.logout()
        self.anon_client.logout()
        shutil.rmtree(self.temp_dir)

    @staticmethod
    def __create_file(target_file, size):
        file_handle = open(target_file, 'w')
        dummy_data = ''
        for i in range(0, 1024):
            dummy_data += 'X'

        for i in range(0, int(size / 1024)):
            # write in 1kb blocks
            file_handle.write(dummy_data)

        dummy_data = ''
        for i in range(0, size % 1024):
            dummy_data += 'X'

        if (len(dummy_data) > 0):
            file_handle.write(dummy_data)

        file_handle.close()

    def test_anon_login(self):
        self.anon_client = nextcloud.Client(Config['owncloud_url'])
        self.anon_client.anon_login(self.public_download_token)
        listing = self.anon_client.list('/')
        self.assertEqual(len(listing), 1)

        # repeat with wrong token
        self.anon_client = nextcloud.Client(Config['owncloud_url'])
        self.anon_client.anon_login('badtoken')
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.anon_client.list('/')
        self.assertEqual(e.exception.status_code, 401)

        # repeat with upload only folder (listing not allowed)
        self.anon_client = nextcloud.Client(Config['owncloud_url'])
        self.anon_client.anon_login(self.public_upload_token)
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.anon_client.list('/')
        self.assertEqual(e.exception.status_code, 404)
    
    def test_anon_login_pw(self):
        self.anon_client = nextcloud.Client(Config['owncloud_url'])
        self.anon_client.anon_login(self.pw_public_download_token, folder_password=self.folder_password)
        listing = self.anon_client.list('/')
        self.assertEqual(len(listing), 1)

        # repeat with wrong password
        self.anon_client = nextcloud.Client(Config['owncloud_url'])
        self.anon_client.anon_login(self.pw_public_download_token, folder_password='wrongpassword')
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.anon_client.list('/')
        self.assertEqual(e.exception.status_code, 401)

        # repeat with upload only folder (listing not allowed)
        self.anon_client = nextcloud.Client(Config['owncloud_url'])
        self.anon_client.anon_login(self.pw_public_upload_token, folder_password=self.folder_password)
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.anon_client.list('/')
        self.assertEqual(e.exception.status_code, 404)

    def test_from_link(self):
        self.anon_client = nextcloud.Client.from_public_link(self.public_download_link)
        listing = self.anon_client.list('/')
        self.assertEqual(len(listing), 1)

        # repeat with wrong link
        self.anon_client = nextcloud.Client.from_public_link(Config['owncloud_url'] + 'index.php/s/wronglink')
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.anon_client.list('/')
        self.assertEqual(e.exception.status_code, 401)

        # repeat with upload only folder (listing not allowed)
        self.anon_client = nextcloud.Client.from_public_link(self.public_upload_link)
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.anon_client.list('/')
        self.assertEqual(e.exception.status_code, 404)
    
    def test_from_link_pw(self):
        self.anon_client = nextcloud.Client.from_public_link(self.pw_public_download_link, folder_password=self.folder_password)
        listing = self.anon_client.list('/')
        self.assertEqual(len(listing), 1)

        # repeat with wrong password
        self.anon_client = nextcloud.Client.from_public_link(self.pw_public_download_link, folder_password='wrongpassword')
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.anon_client.list('/')
        self.assertEqual(e.exception.status_code, 401)

        # repeat with upload only folder (listing not allowed)
        self.anon_client = nextcloud.Client.from_public_link(self.pw_public_upload_link, folder_password=self.folder_password)
        with self.assertRaises(nextcloud.ResponseError) as e:
            self.anon_client.list('/')
        self.assertEqual(e.exception.status_code, 404)             


    @data_provider(files_content)
    def test_download_file(self, file_name, content, subdir):
        """Test file download"""
        # populate folder with data
        temp_file = self.temp_dir + file_name
        self.client.put_file_contents(self.public_folder_download + '/' + file_name, content)

        self.anon_client = nextcloud.Client(Config['owncloud_url'])
        self.anon_client.anon_login(self.public_download_token)

        self.assertTrue(self.anon_client.get_file('/' + file_name, temp_file))

        f = open(temp_file, 'rb')
        s = f.read()
        f.close()
        os.unlink(temp_file)
        self.assertEqual(s, content)

    # ZIP Download requires new method for anon session. API URL differs from regular session.
    # def test_download_dir(self, file_name, content, subdir):

    @data_provider(files_content)
    def test_mkdir(self, file_name, content, subdir):
        """Test subdirectory creation"""
        self.anon_client = nextcloud.Client(Config['owncloud_url'])
        self.anon_client.anon_login(self.public_upload_token)

        self.assertTrue(self.anon_client.mkdir(subdir))
        # Only users with account can get file info in upload folders
        self.assertIsNotNone(self.client.file_info(self.public_folder_upload + subdir))
    
    @data_provider(files)
    def test_upload_small_file(self, file_name):
        """Test simple upload"""
        temp_file = self.temp_dir + 'pyoctest.dat'
        self.__create_file(temp_file, 2 * 1024)
        self.anon_client = nextcloud.Client(Config['owncloud_url'])
        self.anon_client.anon_login(self.public_upload_token)

        self.assertTrue(self.anon_client.put_file('/' + file_name, temp_file, chunked=False))
        os.unlink(temp_file)

        file_info = self.client.file_info(self.public_folder_upload + file_name)
        self.assertIsNotNone(file_info)
        self.assertEqual(file_info.get_size(), 2 * 1024)
    
    @data_provider(files)
    def test_upload_big_file(self, file_name):
        """Test chunked upload"""
        temp_file = self.temp_dir + 'pyoctest.dat'
        self.__create_file(temp_file, 10 * 1024 * 1024)

        self.anon_client = nextcloud.Client(Config['owncloud_url'])
        self.anon_client.anon_login(self.public_upload_token)

        self.assertTrue(self.anon_client.put_file('/' + file_name, temp_file))
        os.unlink(temp_file)

        file_info = self.client.file_info(self.public_folder_upload + file_name)
        self.assertIsNotNone(file_info)
        self.assertEqual(file_info.get_size(), 10 * 1024 * 1024)

    @data_provider(files)
    def test_drop_file(self, file_name):
        """Test convinience wrapper"""
        temp_file = self.temp_dir + file_name
        self.__create_file(temp_file, 2 * 1024)
        self.anon_client = nextcloud.Client(Config['owncloud_url'])
        self.anon_client.anon_login(self.public_upload_token)

        self.assertTrue(self.anon_client.drop_file(temp_file))
        os.unlink(temp_file)

        file_info = self.client.file_info(self.public_folder_upload + file_name)
        self.assertIsNotNone(file_info)
        self.assertEqual(file_info.get_size(), 2 * 1024)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# vim: expandtab shiftwidth=4 softtabstop=4
#
"""nextCloud client module

Makes it possible to access files on a remote nextCloud instance,
share them or access application attributes.
"""

import datetime
import time
import requests
import xml.etree.ElementTree as ET
import os
import math
import six
from six.moves.urllib import parse


class ResponseError(Exception):
    def __init__(self, res, errorType):
        if type(res) is int:
            code = res
        else:
            code = res.status_code
            self.res = res
        Exception.__init__(self, errorType + " error: %i" % code)
        self.status_code = code

    def get_resource_body(self):
        if self.res is not None:
            return self.res.content
        else:
            return None


class OCSResponseError(ResponseError):
    def __init__(self, res):
        ResponseError.__init__(self, res, "OCS")

    def get_resource_body(self):
        if self.res is not None:
            import xml.etree.ElementTree as ElementTree
            try:
                root_element = ElementTree.fromstringlist(self.res.content)
                if root_element.tag == 'message':
                    return root_element.text
            except ET.ParseError:
                return self.res.content
        else:
            return None


class HTTPResponseError(ResponseError):
    def __init__(self, res):
        ResponseError.__init__(self, res, "HTTP")


class ShareInfo(object):
    """Share information"""

    def __init__(self, share_info):
        self.share_info = {}
        # remove unneeded attributes
        del_attrs = ['item_type', 'item_source', 'file_source', 'parent',
                     'storage', 'mail_send']
        for k, v in share_info.items():
            if k not in del_attrs:
                self.share_info[k] = v
        self.share_id = self._get_int('id')
        if 'token' in self.share_info:
            self.token = self.share_info['token']

    def get_id(self):
        """Returns the id of the share

        :returns: id of the share
        """
        return self._get_int('id')

    def get_share_type(self):
        """Returns the type of the share.
        See OCS_SHARE_TYPE_* constants.

        :returns: share type
        """
        return self._get_int('share_type')

    def get_share_with(self):
        """Returns the share recipient.
        If share type is OCS_SHARE_TYPE_USER, then the recipient is the name of
        the user.
        For OCS_SHARE_TYPE_GROUP it is the name of the group.
        Otherwise this value is None.

        :returns: name of the share recipient
        """
        if 'share_with' in self.share_info:
            return self.share_info['share_with']
        return None

    def get_share_with_displayname(self):
        """Returns the share recipient displayname.
        If share_with_displayname cannot be returned, None is returned instead
        :returns: name of the share recipient
        """
        if 'share_with_displayname' in self.share_info:
            return self.share_info['share_with_displayname']
        return None

    def get_path(self):
        """Returns the path of the shared file/folder relative to the
        caller's filesystem.

        :returns: path to the shared file/folder
        """
        if 'path' in self.share_info:
            return self.share_info['path']
        return None

    def get_permissions(self):
        """Returns the share permissions.
        See OCS_PERMISSION_* constants.

        :returns: share permissions
        """
        return self._get_int('permissions')

    def get_share_time(self):
        """Returns the share time.

        :returns: share timestamp
        :rtype: datetime object
        """
        return datetime.datetime.fromtimestamp(
            self._get_int('stime')
        )

    def get_expiration(self):
        """Returns the expiration date.

        :returns: expiration date
        :rtype: datetime object
        """
        exp = self._get_int('expiration')
        if exp is not None:
            return datetime.datetime.fromtimestamp(
                exp
            )
        return None

    def get_token(self):
        if 'token' in self.share_info:
            return self.share_info['token']
        return None

    def get_link(self):
        if 'url' in self.share_info:
            return self.share_info['url']
        return None

    def get_uid_owner(self):
        """Returns the user id of the owner.

        :returns: owner user id
        """
        if 'uid_owner' in self.share_info:
            return self.share_info['uid_owner']
        return None

    def get_displayname_owner(self):
        """Returns the display name of the owner.

        :returns: display name of owner
        """
        if 'displayname_owner' in self.share_info:
            return self.share_info['displayname_owner']
        return None

    def get_name(self):
        """Returns the name of a link share

        :returns: name of link share
        """
        if 'name' in self.share_info:
            return self.share_info['name']
        return None

    def __str__(self):
        info = ''
        for k, v in self.share_info.items():
            info += '%s=%s,' % (k, v)
        return 'ShareInfo(%s)' % info[:-1]

    def __repr__(self):
        return self.__str__()

    def _get_int(self, key):
        """Simple wrapper which converts value to Integer
           in silently manner (w/o raising exception)"""
        try:
            value = int(self.share_info[key])
            return value
        except:
            return None


class PublicShare(ShareInfo):
    """Public share information"""

    def __init__(self, share_id, target_file, link, token):
        self.share_id = share_id
        self.target_file = target_file
        self.link = link
        self.token = token

    def __str__(self):
        return 'PublicShare(id=%i,path=%s,link=%s,token=%s)' % \
               (self.share_id, self.target_file, self.link, self.token)


class FileInfo(object):
    """File information"""

    _DATE_FORMAT = '%a, %d %b %Y %H:%M:%S %Z'

    def __init__(self, path, file_type='file', attributes=None):
        self.path = path
        if path.endswith('/'):
            path = path[0:-1]
        self.name = os.path.basename(path)
        self.file_type = file_type
        self.attributes = attributes or {}

    def get_name(self):
        """Returns the base name of the file without path

        :returns: name of the file
        """
        return self.name

    def get_path(self):
        """Returns the full path to the file without name and without
        trailing slash

        :returns: path to the file
        """
        return os.path.dirname(self.path)

    def get_size(self):
        """Returns the size of the file

        :returns: size of the file
        """
        if '{DAV:}getcontentlength' in self.attributes:
            return int(self.attributes['{DAV:}getcontentlength'])
        return None

    def get_etag(self):
        """Returns the file etag

        :returns: file etag
        """
        return self.attributes['{DAV:}getetag']

    def get_content_type(self):
        """Returns the file content type

        :returns: file content type
        """
        if '{DAV:}getcontenttype' in self.attributes:
            return self.attributes['{DAV:}getcontenttype']

        if self.is_dir():
            return 'httpd/unix-directory'

        return None

    def get_last_modified(self):
        """Returns the last modified time

        :returns: last modified time
        :rtype: datetime object
        """
        return datetime.datetime.strptime(
            self.attributes['{DAV:}getlastmodified'],
            self._DATE_FORMAT
        )

    def is_dir(self):
        """Returns whether the file info is a directory

        :returns: True if it is a directory, False otherwise
        """
        return self.file_type != 'file'

    def __str__(self):
        return 'File(path=%s,file_type=%s,attributes=%s)' % \
               (self.path, self.file_type, self.attributes)

    def __repr__(self):
        return self.__str__()


class Client(object):
    """nextCloud client"""

    OCS_BASEPATH = 'ocs/v1.php/'
    OCS_SERVICE_SHARE = 'apps/files_sharing/api/v1'
    OCS_SERVICE_PRIVATEDATA = 'privatedata'
    OCS_SERVICE_CLOUD = 'cloud'

    # constants from lib/public/constants.php
    OCS_PERMISSION_READ = 1
    OCS_PERMISSION_UPDATE = 2
    OCS_PERMISSION_CREATE = 4
    OCS_PERMISSION_DELETE = 8
    OCS_PERMISSION_SHARE = 16
    OCS_PERMISSION_ALL = 31
    # constants from lib/public/share.php
    OCS_SHARE_TYPE_USER = 0
    OCS_SHARE_TYPE_GROUP = 1
    OCS_SHARE_TYPE_LINK = 3
    OCS_SHARE_TYPE_REMOTE = 6

    def __init__(self, url, **kwargs):
        """Instantiates a client

        :param url: URL of the target nextCloud instance
        :param verify_certs: True (default) to verify SSL certificates, False otherwise
        :param dav_endpoint_version: None (default) to force using a specific endpoint version
        instead of relying on capabilities
        :param debug: set to True to print debugging messages to stdout, defaults to False
        """
        if not url.endswith('/'):
            url += '/'

        self.url = url
        self._session = None
        self._debug = kwargs.get('debug', False)
        self._verify_certs = kwargs.get('verify_certs', True)
        self._dav_endpoint_version = kwargs.get('dav_endpoint_version', True)

        self._capabilities = None
        self._version = None

    def login(self, user_id, password):
        """Authenticate to nextCloud.
        This will create a session on the server.

        :param user_id: user id
        :param password: password
        :raises: HTTPResponseError in case an HTTP error status was returned
        """

        self._session = requests.session()
        self._session.verify = self._verify_certs
        self._session.auth = (user_id, password)

        try:
            self._update_capabilities()

            url_components = parse.urlparse(self.url)
            if self._dav_endpoint_version == 1:
                self._davpath = url_components.path + 'remote.php/dav/files/' + parse.quote(user_id)
                self._webdav_url = self.url + 'remote.php/dav/files/' + parse.quote(user_id)
            else:
                self._davpath = url_components.path + 'remote.php/webdav'
                self._webdav_url = self.url + 'remote.php/webdav'

        except HTTPResponseError as e:
            self._session.close()
            self._session = None
            raise e

    def logout(self):
        """Log out the authenticated user and close the session.

        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        # TODO actual logout ?
        self._session.close()
        return True

    def anon_login(self, folder_token, folder_password=''):
        self._session = requests.session()
        self._session.verify = self._verify_certs
        self._session.auth = (folder_token, folder_password)

        url_components = parse.urlparse(self.url)
        self._davpath = url_components.path + 'public.php/webdav'
        self._webdav_url = self.url + 'public.php/webdav'
    
    @classmethod
    def from_public_link(cls, public_link, folder_password='', **kwargs):
        public_link_components = parse.urlparse(public_link)
        url = public_link_components.scheme + '://' + public_link_components.hostname
        folder_token = public_link_components.path.split('/')[-1]       
        anon_session = cls(url, **kwargs)
        anon_session.anon_login(folder_token, folder_password=folder_password)
        return anon_session

    def drop_file(self, file_name):
        """ Convenience wrapper for put_file """
        destination = '/' + os.path.basename(file_name)
        return self.put_file(destination, file_name)

    def file_info(self, path, properties=None):
        """Returns the file info for the given remote file

        :param path: path to the remote file
        :param properties: a list of properties to request (optional)
        :returns: file info
        :rtype: :class:`FileInfo` object or `None` if file
            was not found
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        if properties:
            root = ET.Element('d:propfind',
                              {
                                  'xmlns:d': "DAV:",
                                  'xmlns:nc': "http://nextcloud.org/ns",
                                  'xmlns:oc': "http://nextcloud.org/ns"
                              })
            prop = ET.SubElement(root, 'd:prop')
            for p in properties:
                ET.SubElement(prop, p)
            data = ET.tostring(root)
        else:
            data = None
        res = self._make_dav_request('PROPFIND', path, headers={'Depth': '0'}, data=data)
        if res:
            return res[0]
        return None

    def list(self, path, depth=1, properties=None):
        """Returns the listing/contents of the given remote directory

        :param path: path to the remote directory
        :param depth: depth of the listing, integer or "infinity"
        :param properties: a list of properties to request (optional)
        :returns: directory listing
        :rtype: array of :class:`FileInfo` objects
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        if not path.endswith('/'):
            path += '/'

        headers = {}
        if isinstance(depth, int) or depth == "infinity":
            headers['Depth'] = str(depth)

        if properties:
            root = ET.Element('d:propfind',
                              {
                                  'xmlns:d': "DAV:",
                                  'xmlns:nc': "http://nextcloud.org/ns",
                                  'xmlns:oc': "http://nextcloud.org/ns"
                              })
            prop = ET.SubElement(root, 'd:prop')
            for p in properties:
                ET.SubElement(prop, p)
            data = ET.tostring(root)
        else:
            data = None

        res = self._make_dav_request('PROPFIND', path, headers=headers, data=data)
        # first one is always the root, remove it from listing
        if res:
            return res[1:]
        return None

    def get_file_contents(self, path):
        """Returns the contents of a remote file

        :param path: path to the remote file
        :returns: file contents
        :rtype: binary data
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        path = self._normalize_path(path)
        res = self._session.get(
            self._webdav_url + parse.quote(self._encode_string(path))
        )
        if res.status_code == 200:
            return res.content
        elif res.status_code >= 400:
            raise HTTPResponseError(res)
        return False

    def get_file(self, remote_path, local_file=None):
        """Downloads a remote file

        :param remote_path: path to the remote file
        :param local_file: optional path to the local file. If none specified,
            the file will be downloaded into the current directory
        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        remote_path = self._normalize_path(remote_path)
        res = self._session.get(
            self._webdav_url + parse.quote(self._encode_string(remote_path)),
            stream=True
        )
        if res.status_code == 200:
            if local_file is None:
                # use downloaded file name from Content-Disposition
                # local_file = res.headers['content-disposition']
                local_file = os.path.basename(remote_path)

            file_handle = open(local_file, 'wb', 8192)
            for chunk in res.iter_content(8192):
                file_handle.write(chunk)
            file_handle.close()
            return True
        elif res.status_code >= 400:
            raise HTTPResponseError(res)
        return False

    def get_directory_as_zip(self, remote_path, local_file):
        """Downloads a remote directory as zip

        :param remote_path: path to the remote directory to download
        :param local_file: path and name of the target local file
        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        remote_path = self._normalize_path(remote_path)
        url = self.url + 'index.php/apps/files/ajax/download.php?dir=' \
              + parse.quote(remote_path)
        res = self._session.get(url, stream=True)
        if res.status_code == 200:
            if local_file is None:
                # use downloaded file name from Content-Disposition
                # targetFile = res.headers['content-disposition']
                local_file = os.path.basename(remote_path)

            file_handle = open(local_file, 'wb', 8192)
            for chunk in res.iter_content(8192):
                file_handle.write(chunk)
            file_handle.close()
            return True
        elif res.status_code >= 400:
            raise HTTPResponseError(res)
        return False

    def put_file_contents(self, remote_path, data):
        """Write data into a remote file

        :param remote_path: path of the remote file
        :param data: data to write into the remote file
        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        return self._make_dav_request('PUT', remote_path, data=data)

    def put_file(self, remote_path, local_source_file, **kwargs):
        """Upload a file

        :param remote_path: path to the target file. A target directory can
            also be specified instead by appending a "/"
        :param local_source_file: path to the local file to upload
        :param chunked: (optional) use file chunking (defaults to True)
        :param chunk_size: (optional) chunk size in bytes, defaults to 10 MB
        :param keep_mtime: (optional) also update the remote file to the same
            mtime as the local one, defaults to True
        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        if kwargs.get('chunked', True):
            return self._put_file_chunked(
                remote_path,
                local_source_file,
                **kwargs
            )

        stat_result = os.stat(local_source_file)

        headers = {}
        if kwargs.get('keep_mtime', True):
            headers['X-OC-MTIME'] = str(int(stat_result.st_mtime))

        if remote_path[-1] == '/':
            remote_path += os.path.basename(local_source_file)
        file_handle = open(local_source_file, 'rb', 8192)
        res = self._make_dav_request(
            'PUT',
            remote_path,
            data=file_handle,
            headers=headers
        )
        file_handle.close()
        return res

    def put_directory(self, target_path, local_directory, **kwargs):
        """Upload a directory with all its contents

        :param target_path: path of the directory to upload into
        :param local_directory: path to the local directory to upload
        :param \*\*kwargs: optional arguments that ``put_file`` accepts
        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        target_path = self._normalize_path(target_path)
        if not target_path.endswith('/'):
            target_path += '/'
        gathered_files = []

        if not local_directory.endswith('/'):
            local_directory += '/'

        basedir = os.path.basename(local_directory[0: -1]) + '/'
        # gather files to upload
        for path, _, files in os.walk(local_directory):
            gathered_files.append(
                (path, basedir + path[len(local_directory):], files)
            )

        for path, remote_path, files in gathered_files:
            self.mkdir(target_path + remote_path + '/')
            for name in files:
                if not self.put_file(target_path + remote_path + '/',
                                     path + '/' + name, **kwargs):
                    return False
        return True

    def _put_file_chunked(self, remote_path, local_source_file, **kwargs):
        """Uploads a file using chunks. If the file is smaller than
        ``chunk_size`` it will be uploaded directly.

        :param remote_path: path to the target file. A target directory can
        also be specified instead by appending a "/"
        :param local_source_file: path to the local file to upload
        :param \*\*kwargs: optional arguments that ``put_file`` accepts
        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        chunk_size = kwargs.get('chunk_size', 10 * 1024 * 1024)
        result = True
        transfer_id = int(time.time())

        remote_path = self._normalize_path(remote_path)
        if remote_path.endswith('/'):
            remote_path += os.path.basename(local_source_file)

        stat_result = os.stat(local_source_file)

        file_handle = open(local_source_file, 'rb', 8192)
        file_handle.seek(0, os.SEEK_END)
        size = file_handle.tell()
        file_handle.seek(0)

        headers = {}
        if kwargs.get('keep_mtime', True):
            headers['X-OC-MTIME'] = str(int(stat_result.st_mtime))

        if size == 0:
            return self._make_dav_request(
                'PUT',
                remote_path,
                data='',
                headers=headers
            )

        chunk_count = int(math.ceil(float(size) / float(chunk_size)))

        if chunk_count > 1:
            headers['OC-CHUNKED'] = '1'

        for chunk_index in range(0, int(chunk_count)):
            data = file_handle.read(chunk_size)
            if chunk_count > 1:
                chunk_name = '%s-chunking-%s-%i-%i' % \
                             (remote_path, transfer_id, chunk_count,
                              chunk_index)
            else:
                chunk_name = remote_path

            if not self._make_dav_request(
                    'PUT',
                    chunk_name,
                    data=data,
                    headers=headers
            ):
                result = False
                break

        file_handle.close()
        return result

    def mkdir(self, path):
        """Creates a remote directory

        :param path: path to the remote directory to create
        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        if not path.endswith('/'):
            path += '/'
        return self._make_dav_request('MKCOL', path)

    def delete(self, path):
        """Deletes a remote file or directory

        :param path: path to the file or directory to delete
        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        return self._make_dav_request('DELETE', path)

    def list_open_remote_share(self):
        """List all pending remote shares

        :returns: array of pending remote shares
        :raises: HTTPResponseError in case an HTTP error status was returned
        """

        res = self._make_ocs_request(
            'GET',
            self.OCS_SERVICE_SHARE,
            'remote_shares/pending'
        )
        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree)
            shares = []
            for element in tree.find('data').iter('element'):
                share_attr = {}
                for child in element:
                    key = child.tag
                    value = child.text
                    share_attr[key] = value
                shares.append(share_attr)
            return shares
        raise HTTPResponseError(res)

    def accept_remote_share(self, share_id):
        """Accepts a remote share

        :param share_id: Share ID (int)
        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        if not isinstance(share_id, int):
            return False

        res = self._make_ocs_request(
            'POST',
            self.OCS_SERVICE_SHARE,
            'remote_shares/pending/' + str(share_id)
        )
        if res.status_code == 200:
            return res
        raise HTTPResponseError(res)

    def decline_remote_share(self, share_id):
        """Declines a remote share

        :param share_id: Share ID (int)
        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        if not isinstance(share_id, int):
            return False

        res = self._make_ocs_request(
            'DELETE',
            self.OCS_SERVICE_SHARE,
            'remote_shares/pending/' + str(share_id)
        )
        if res.status_code == 200:
            return res
        raise HTTPResponseError(res)

    def delete_share(self, share_id):
        """Unshares a file or directory

        :param share_id: Share ID (int)
        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        if not isinstance(share_id, int):
            return False

        res = self._make_ocs_request(
            'DELETE',
            self.OCS_SERVICE_SHARE,
            'shares/' + str(share_id)
        )
        if res.status_code == 200:
            return res
        raise HTTPResponseError(res)

    def update_share(self, share_id, **kwargs):
        """Updates a given share

        :param share_id: (int) Share ID
        :param perms: (int) update permissions (see share_file_with_user() below)
        :param password: (string) updated password for public link Share
        :param public_upload: (boolean) enable/disable public upload for public shares
        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned
        """

        perms = kwargs.get('perms', None)
        password = kwargs.get('password', None)
        public_upload = kwargs.get('public_upload', None)
        if (isinstance(perms, int)) and (perms > self.OCS_PERMISSION_ALL):
            perms = None
        if not (perms or password or (public_upload is not None)):
            return False
        if not isinstance(share_id, int):
            return False

        data = {}
        if perms:
            data['permissions'] = perms
        if isinstance(password, six.string_types):
            data['password'] = password
        if (public_upload is not None) and (isinstance(public_upload, bool)):
            data['publicUpload'] = str(public_upload).lower()

        res = self._make_ocs_request(
            'PUT',
            self.OCS_SERVICE_SHARE,
            'shares/' + str(share_id),
            data=data
        )
        if res.status_code == 200:
            return True
        raise HTTPResponseError(res)

    def move(self, remote_path_source, remote_path_target):
        """Moves a remote file or directory

        :param remote_path_source: source file or folder to move
        :param remote_path_target: target file to which to move
        the source file. A target directory can also be specified
        instead by appending a "/"
        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned
        """

        return self._webdav_move_copy(remote_path_source, remote_path_target,
                                       "MOVE")

    def copy(self, remote_path_source, remote_path_target):
        """Copies a remote file or directory

        :param remote_path_source: source file or folder to copy
        :param remote_path_target: target file to which to copy

        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        return self._webdav_move_copy(remote_path_source, remote_path_target,
                                       "COPY")

    def share_file_with_link(self, path, **kwargs):
        """Shares a remote file with link

        :param path: path to the remote file to share
        :param perms (optional): permission of the shared object
        defaults to read only (1)
        :param public_upload (optional): allows users to upload files or folders
        :param password (optional): sets a password
        https://docs.nextcloud.com/server/latest/developer_manual/client_apis/OCS/ocs-share-api.html
        :param name (optional): display name for the link
        :returns: instance of :class:`ShareInfo` with the share info
            or False if the operation failed
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        perms = kwargs.get('perms', None)
        public_upload = kwargs.get('public_upload', 'false')
        password = kwargs.get('password', None)
        name = kwargs.get('name', None)

        path = self._normalize_path(path)
        post_data = {
            'shareType': self.OCS_SHARE_TYPE_LINK,
            'path': self._encode_string(path),
        }
        if (public_upload is not None) and (isinstance(public_upload, bool)):
            post_data['publicUpload'] = str(public_upload).lower()
        if isinstance(password, six.string_types):
            post_data['password'] = password
        if name is not None:
            post_data['name'] = self._encode_string(name)
        if perms:
            post_data['permissions'] = perms

        res = self._make_ocs_request(
            'POST',
            self.OCS_SERVICE_SHARE,
            'shares',
            data=post_data
        )
        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree)
            data_el = tree.find('data')
            return ShareInfo(
                                {
                                    'id': data_el.find('id').text,
                                    'path': path,
                                    'url': data_el.find('url').text,
                                    'token': data_el.find('token').text,
                                }
            )
        raise HTTPResponseError(res)

    def is_shared(self, path):
        """Checks whether a path is already shared

        :param path: path to the share to be checked
        :returns: True if the path is already shared, else False
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        # make sure that the path exist - if not, raise HTTPResponseError
        self.file_info(path)
        try:
            result = self.get_shares(path)
            if result:
                return len(result) > 0
        except OCSResponseError as e:
            if e.status_code != 404:
                raise e
            return False
        return False

    def get_share(self, share_id):
        """Returns share information about known share

        :param share_id: id of the share to be checked
        :returns: instance of ShareInfo class
        :raises: ResponseError in case an HTTP error status was returned
        """
        if (share_id is None) or not (isinstance(share_id, int)):
            return None

        res = self._make_ocs_request(
                'GET',
                self.OCS_SERVICE_SHARE,
                'shares/' + str(share_id)
                )
        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree)
            return self._get_shareinfo(tree.find('data').find('element'))
        raise HTTPResponseError(res)

    def get_shares(self, path='', **kwargs):
        """Returns array of shares

        :param path: path to the share to be checked
        :param reshares: (optional, boolean) returns not only the shares from
            the current user but all shares from the given file (default: False)
        :param subfiles: (optional, boolean) returns all shares within
            a folder, given that path defines a folder (default: False)
        :param shared_with_me: (optional, boolean) returns all shares which are
            shared with me (default: False)
        :returns: array of shares ShareInfo instances or empty array if the operation failed
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        if not (isinstance(path, six.string_types)):
            return None

        data = 'shares'
        if path != '':
            data += '?'
            path = self._encode_string(self._normalize_path(path))
            args = {'path': path}
            reshares = kwargs.get('reshares', False)
            if isinstance(reshares, bool) and reshares:
                args['reshares'] = reshares
            subfiles = kwargs.get('subfiles', False)
            if isinstance(subfiles, bool) and subfiles:
                args['subfiles'] = str(subfiles).lower()

            shared_with_me = kwargs.get('shared_with_me', False)
            if isinstance(shared_with_me, bool) and shared_with_me:
                args['shared_with_me'] = "true"
                del args['path']

            data += parse.urlencode(args)

        res = self._make_ocs_request(
            'GET',
            self.OCS_SERVICE_SHARE,
            data
        )

        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree)
            shares = []
            for element in tree.find('data').iter('element'):
                '''share_attr = {}
                for child in element:
                    key = child.tag
                    value = child.text
                    share_attr[key] = value
                shares.append(share_attr)'''
                shares.append(self._get_shareinfo(element))
            return shares
        raise HTTPResponseError(res)

    def create_user(self, user_name, initial_password):
        """Create a new user with an initial password via provisioning API.
        It is not an error, if the user already existed before.
        If you get back an error 999, then the provisioning API is not enabled.

        :param user_name:  name of user to be created
        :param initial_password:  password for user being created
        :returns: True on success
        :raises: HTTPResponseError in case an HTTP error status was returned

        """
        res = self._make_ocs_request(
            'POST',
            self.OCS_SERVICE_CLOUD,
            'users',
            data={'password': initial_password, 'userid': user_name}
        )

        # We get 200 when the user was just created.
        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree, [100])
            return True

        raise HTTPResponseError(res)

    def delete_user(self, user_name):
        """Deletes a user via provisioning API.
        If you get back an error 999, then the provisioning API is not enabled.

        :param user_name:  name of user to be deleted
        :returns: True on success
        :raises: HTTPResponseError in case an HTTP error status was returned

        """
        res = self._make_ocs_request(
            'DELETE',
            self.OCS_SERVICE_CLOUD,
            'users/' + user_name
        )

        # We get 200 when the user was deleted.
        if res.status_code == 200:
            return True

        raise HTTPResponseError(res)

    def user_exists(self, user_name):
        """Checks a user via provisioning API.
        If you get back an error 999, then the provisioning API is not enabled.

        :param user_name:  name of user to be checked
        :returns: True if user found

        """
        users = self.search_users(user_name)

        return user_name in users

    def search_users(self, user_name):
        """Searches for users via provisioning API.
        If you get back an error 999, then the provisioning API is not enabled.

        :param user_name:  name of user to be searched for
        :returns: list of usernames that contain user_name as substring
        :raises: HTTPResponseError in case an HTTP error status was returned

        """
        action_path = 'users'
        if user_name:
            action_path += '?search={}'.format(user_name)

        res = self._make_ocs_request(
            'GET',
            self.OCS_SERVICE_CLOUD,
            action_path
        )

        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            users = [x.text for x in tree.findall('data/users/element')]

            return users

        raise HTTPResponseError(res)

    def get_users(self):
        """Get users via provisioning API.
        If you get back an error 999, then the provisioning API is not enabled.

        :returns: list of usernames
        :raises: HTTPResponseError in case an HTTP error status was returned

        """
        return self.search_users('')

    def set_user_attribute(self, user_name, key, value):
        """Sets a user attribute

        :param user_name: name of user to modify
        :param key: key of the attribute to set
        :param value: value to set
        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned
        """

        res = self._make_ocs_request(
            'PUT',
            self.OCS_SERVICE_CLOUD,
            'users/' + parse.quote(user_name),
            data={'key': self._encode_string(key),
                  'value': self._encode_string(value)}
        )

        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree, [100])
            return True
        raise HTTPResponseError(res)

    def add_user_to_group(self, user_name, group_name):
        """Adds a user to a group.

        :param user_name:  name of user to be added
        :param group_name:  name of group user is to be added to
        :returns: True if user added
        :raises: HTTPResponseError in case an HTTP error status was returned

        """

        res = self._make_ocs_request(
            'POST',
            self.OCS_SERVICE_CLOUD,
            'users/' + user_name + '/groups',
            data={'groupid': group_name}
        )

        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree, [100])
            return True

        raise HTTPResponseError(res)

    def get_user_groups(self, user_name):
        """Get a list of groups associated to a user.

        :param user_name:  name of user to list groups
        :returns: list of groups
        :raises: HTTPResponseError in case an HTTP error status was returned

        """

        res = self._make_ocs_request(
            'GET',
            self.OCS_SERVICE_CLOUD,
            'users/' + user_name + '/groups',
        )

        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree, [100])
            return [group.text for group in tree.find('data/groups')]

        raise HTTPResponseError(res)

    def user_is_in_group(self, user_name, group_name):
        """Checks if a user is in a group

        :param user_name:  name of user
        :param group_name:  name of group
        :returns: True if user is in group
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        return group_name in self.get_user_groups(user_name)

    def get_user(self, user_name):
        """Retrieves information about a user

        :param user_name:  name of user to query

        :returns: Dictionary of information about user
        :raises: ResponseError in case an HTTP error status was returned
        """
        res = self._make_ocs_request(
            'GET',
            self.OCS_SERVICE_CLOUD,
            'users/' + parse.quote(user_name),
            data={}
        )

        tree = ET.fromstring(res.content)
        self._check_ocs_status(tree)
        # <ocs><meta><statuscode>100</statuscode><status>ok</status></meta>
        # <data>
        # <email>frank@example.org</email><quota>0</quota><enabled>true</enabled>
        # </data>
        # </ocs>

        data_element = tree.find('data')
        return self._xml_to_dict(data_element)

    def remove_user_from_group(self, user_name, group_name):
        """Removes a user from a group.

        :param user_name:  name of user to be removed
        :param group_name:  name of group user is to be removed from
        :returns: True if user removed
        :raises: HTTPResponseError in case an HTTP error status was returned

        """
        res = self._make_ocs_request(
            'DELETE',
            self.OCS_SERVICE_CLOUD,
            'users/' + user_name + '/groups',
            data={'groupid': group_name}
        )

        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree, [100])
            return True

        raise HTTPResponseError(res)

    def add_user_to_subadmin_group(self, user_name, group_name):
        """Adds a user to a subadmin group.

        :param user_name:  name of user to be added to subadmin group
        :param group_name:  name of subadmin group
        :returns: True if user added
        :raises: HTTPResponseError in case an HTTP error status was returned

        """

        res = self._make_ocs_request(
            'POST',
            self.OCS_SERVICE_CLOUD,
            'users/' + user_name + '/subadmins',
            data={'groupid': group_name}
        )

        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree, [100, 103])
            return True

        raise HTTPResponseError(res)

    def get_user_subadmin_groups(self, user_name):
        """Get a list of subadmin groups associated to a user.

        :param user_name:  name of user
        :returns: list of subadmin groups
        :raises: HTTPResponseError in case an HTTP error status was returned

        """

        res = self._make_ocs_request(
            'GET',
            self.OCS_SERVICE_CLOUD,
            'users/' + user_name + '/subadmins',
        )

        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree, [100])

            groups = tree.find('data')

            return groups

        raise HTTPResponseError(res)

    def user_is_in_subadmin_group(self, user_name, group_name):
        """Checks if a user is in a subadmin group

        :param user_name:  name of user
        :param group_name:  name of subadmin group
        :returns: True if user is in subadmin group
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        return group_name in self.get_user_subadmin_groups(user_name)

    def share_file_with_user(self, path, user, **kwargs):
        """Shares a remote file with specified user

        :param path: path to the remote file to share
        :param user: name of the user whom we want to share a file/folder
        :param perms (optional): permissions of the shared object
            defaults to read only (1)
            https://docs.nextcloud.com/server/latest/developer_manual/client_apis/OCS/ocs-share-api.html
        :param remote_user (optional): True if it is a federated users
            defaults to False if it is a local user
        :returns: instance of :class:`ShareInfo` with the share info
            or False if the operation failed
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        remote_user = kwargs.get('remote_user', False)
        perms = kwargs.get('perms', self.OCS_PERMISSION_READ)
        if (((not isinstance(perms, int)) or (perms > self.OCS_PERMISSION_ALL))
                or ((not isinstance(user, six.string_types)) or (user == ''))):
            return False

        if remote_user and (not user.endswith('/')):
            user = user + '/'
        path = self._normalize_path(path)
        post_data = {
            'shareType': self.OCS_SHARE_TYPE_REMOTE if remote_user else
            self.OCS_SHARE_TYPE_USER,
            'shareWith': user,
            'path': self._encode_string(path),
            'permissions': perms
        }

        res = self._make_ocs_request(
            'POST',
            self.OCS_SERVICE_SHARE,
            'shares',
            data=post_data
        )

        if self._debug:
            print('OCS share_file request for file %s with permissions %i '
                  'returned: %i' % (path, perms, res.status_code))
        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree)
            data_el = tree.find('data')
            return ShareInfo(
                                {
                                    'id': data_el.find('id').text,
                                    'path': path,
                                    'permissions': perms
                                }
            )
        raise HTTPResponseError(res)

    def create_group(self, group_name):
        """Create a new group via provisioning API.
        If you get back an error 999, then the provisioning API is not enabled.

        :param group_name:  name of group to be created
        :returns: True if group created
        :raises: HTTPResponseError in case an HTTP error status was returned

        """
        res = self._make_ocs_request(
            'POST',
            self.OCS_SERVICE_CLOUD,
            'groups',
            data={'groupid': group_name}
        )

        # We get 200 when the group was just created.
        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree, [100])
            return True

        raise HTTPResponseError(res)

    def delete_group(self, group_name):
        """Delete a group via provisioning API.
        If you get back an error 999, then the provisioning API is not enabled.

        :param group_name:  name of group to be deleted
        :returns: True if group deleted
        :raises: HTTPResponseError in case an HTTP error status was returned

        """
        res = self._make_ocs_request(
            'DELETE',
            self.OCS_SERVICE_CLOUD,
            'groups/' + group_name
        )

        # We get 200 when the group was just deleted.
        if res.status_code == 200:
            return True

        raise HTTPResponseError(res)

    def get_groups(self):
        """Get groups via provisioning API.
        If you get back an error 999, then the provisioning API is not enabled.

        :returns: list of groups
        :raises: HTTPResponseError in case an HTTP error status was returned

        """
        res = self._make_ocs_request(
            'GET',
            self.OCS_SERVICE_CLOUD,
            'groups'
        )

        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            groups = [x.text for x in tree.findall('data/groups/element')]

            return groups

        raise HTTPResponseError(res)

    def get_group_members(self, group_name):
        """Get group members via provisioning API.
        If you get back an error 999, then the provisioning API is not enabled.

        :param group_name:  name of group to list members
        :returns: list of group members
        :raises: HTTPResponseError in case an HTTP error status was returned

        """
        res = self._make_ocs_request(
            'GET',
            self.OCS_SERVICE_CLOUD,
            'groups/' + group_name
        )

        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree, [100])
            return [group.text for group in tree.find('data/users')]

        raise HTTPResponseError(res)

    def group_exists(self, group_name):
        """Checks a group via provisioning API.
        If you get back an error 999, then the provisioning API is not enabled.

        :param group_name:  name of group to be checked
        :returns: True if group exists
        :raises: HTTPResponseError in case an HTTP error status was returned

        """
        res = self._make_ocs_request(
            'GET',
            self.OCS_SERVICE_CLOUD,
            'groups?search=' + group_name
        )

        if res.status_code == 200:
            tree = ET.fromstring(res.content)

            for code_el in tree.findall('data/groups/element'):
                if code_el is not None and code_el.text == group_name:
                    return True

            return False

        raise HTTPResponseError(res)

    def share_file_with_group(self, path, group, **kwargs):
        """Shares a remote file with specified group

        :param path: path to the remote file to share
        :param group: name of the group with which we want to share a file/folder
        :param perms (optional): permissions of the shared object
            defaults to read only (1)
            https://docs.nextcloud.com/server/latest/developer_manual/client_apis/OCS/ocs-share-api.html
        :returns: instance of :class:`ShareInfo` with the share info
            or False if the operation failed
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        perms = kwargs.get('perms', self.OCS_PERMISSION_READ)
        if (((not isinstance(perms, int)) or (perms > self.OCS_PERMISSION_ALL))
                or ((not isinstance(group, six.string_types)) or (group == ''))):
            return False

        path = self._normalize_path(path)
        post_data = {'shareType': self.OCS_SHARE_TYPE_GROUP,
                     'shareWith': group,
                     'path': path,
                     'permissions': perms}

        res = self._make_ocs_request(
            'POST',
            self.OCS_SERVICE_SHARE,
            'shares',
            data=post_data
        )
        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree)
            data_el = tree.find('data')
            return ShareInfo(
                                {
                                    'id': data_el.find('id').text,
                                    'path': path,
                                    'permissions': perms
                                }
            )
        raise HTTPResponseError(res)

    def get_config(self):
        """Returns nextCloud config information
        :returns: array of tuples (key, value) for each information
            e.g. [('version', '1.7'), ('website', 'nextCloud'), ('host', 'cloud.example.com'),
            ('contact', ''), ('ssl', 'false')]
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        path = 'config'
        res = self._make_ocs_request(
            'GET',
            '',
            path
        )
        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree)
            values = []

            element = tree.find('data')
            if element is not None:
                keys = ['version', 'website', 'host', 'contact', 'ssl']
                for key in keys:
                    text = element.find(key).text or ''
                    values.append(text)
                return zip(keys, values)
            else:
                return None
        raise HTTPResponseError(res)

    def get_attribute(self, app=None, key=None):
        """Returns an application attribute

        :param app: application id
        :param key: attribute key or None to retrieve all values for the
            given application
        :returns: attribute value if key was specified, or an array of tuples
            (key, value) for each attribute
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        path = 'getattribute'
        if app is not None:
            path += '/' + parse.quote(app, '')
            if key is not None:
                path += '/' + parse.quote(self._encode_string(key), '')
        res = self._make_ocs_request(
            'GET',
            self.OCS_SERVICE_PRIVATEDATA,
            path
        )
        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree)
            values = []
            for element in tree.find('data').iter('element'):
                app_text = element.find('app').text
                key_text = element.find('key').text
                value_text = element.find('value').text or ''
                if key is None:
                    if app is None:
                        values.append((app_text, key_text, value_text))
                    else:
                        values.append((key_text, value_text))
                else:
                    return value_text

            if len(values) == 0 and key is not None:
                return None
            return values
        raise HTTPResponseError(res)

    def set_attribute(self, app, key, value):
        """Sets an application attribute

        :param app: application id
        :param key: key of the attribute to set
        :param value: value to set
        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        path = 'setattribute/' + parse.quote(app, '') + '/' + parse.quote(
            self._encode_string(key), '')
        res = self._make_ocs_request(
            'POST',
            self.OCS_SERVICE_PRIVATEDATA,
            path,
            data={'value': self._encode_string(value)}
        )
        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree)
            return True
        raise HTTPResponseError(res)

    def delete_attribute(self, app, key):
        """Deletes an application attribute

        :param app: application id
        :param key: key of the attribute to delete
        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        path = 'deleteattribute/' + parse.quote(app, '') + '/' + parse.quote(
            self._encode_string(key), '')
        res = self._make_ocs_request(
            'POST',
            self.OCS_SERVICE_PRIVATEDATA,
            path
        )
        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree)
            return True
        raise HTTPResponseError(res)

    def get_apps(self):
        """ List all enabled apps through the provisioning api.

        :returns: a dict of apps, with values True/False, representing the enabled state.
        :raises: HTTPResponseError in case an HTTP error status was returned
        """
        ena_apps = {}

        res = self._make_ocs_request('GET', self.OCS_SERVICE_CLOUD, 'apps')
        if res.status_code != 200:
            raise HTTPResponseError(res)
        tree = ET.fromstring(res.content)
        self._check_ocs_status(tree)
        # <data><apps><element>files</element><element>activity</element> ...
        for el in tree.findall('data/apps/element'):
            ena_apps[el.text] = False

        res = self._make_ocs_request('GET', self.OCS_SERVICE_CLOUD,
                                     'apps?filter=enabled')
        if res.status_code != 200:
            raise HTTPResponseError(res)
        tree = ET.fromstring(res.content)
        self._check_ocs_status(tree)
        for el in tree.findall('data/apps/element'):
            ena_apps[el.text] = True

        return ena_apps

    def get_version(self):
        """Gets the nextCloud version of the connected server

        :returns: nextCloud version as string
        """
        if self._version is None:
            self._update_capabilities()
        return self._version

    def get_capabilities(self):
        """Gets the nextCloud app capabilities

        :returns: capabilities dictionary that maps from
        app name to another dictionary containing the capabilities
        """
        if self._capabilities is None:
            self._update_capabilities()
        return self._capabilities

    def enable_app(self, appname):
        """Enable an app through provisioning_api

        :param appname:  Name of app to be enabled
        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned

        """
        res = self._make_ocs_request('POST', self.OCS_SERVICE_CLOUD,
                                     'apps/' + appname)
        if res.status_code == 200:
            return True

        raise HTTPResponseError(res)

    def disable_app(self, appname):
        """Disable an app through provisioning_api

        :param appname:  Name of app to be disabled
        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned

        """
        res = self._make_ocs_request('DELETE', self.OCS_SERVICE_CLOUD,
                                     'apps/' + appname)
        if res.status_code == 200:
            return True

        raise HTTPResponseError(res)

    @staticmethod
    def _normalize_path(path):
        """Makes sure the path starts with a "/"
        """
        if isinstance(path, FileInfo):
            path = path.path
        if len(path) == 0:
            return '/'
        if not path.startswith('/'):
            path = '/' + path
        return path

    @staticmethod
    def _encode_string(s):
        """Encodes a unicode instance to utf-8. If a str is passed it will
        simply be returned

        :param s: str or unicode to encode
        :returns: encoded output as str
        """
        if six.PY2 and isinstance(s, unicode):  # noqa: F821
            return s.encode('utf-8')
        return s

    @staticmethod
    def _check_ocs_status(tree, accepted_codes=[100]):
        """Checks the status code of an OCS request

        :param tree: response parsed with elementtree
        :param accepted_codes: list of statuscodes we consider good. E.g. [100,102] can be used to accept a POST
               returning an 'already exists' condition
        :raises: HTTPResponseError if the http status is not 200, or OCSResponseError if the OCS status is not one of the accepted_codes.
        """
        code_el = tree.find('meta/statuscode')
        if code_el is not None and int(code_el.text) not in accepted_codes:
            r = requests.Response()
            msg_el = tree.find('meta/message')
            if msg_el is None:
                msg_el = tree  # fallback to the entire ocs response, if we find no message.
            r._content = ET.tostring(msg_el)
            r.status_code = int(code_el.text)
            raise OCSResponseError(r)

    def make_ocs_request(self, method, service, action, **kwargs):
        """Makes a OCS API request and analyses the response

        :param method: HTTP method
        :param service: service name
        :param action: action path
        :param \*\*kwargs: optional arguments that ``requests.Request.request`` accepts
        :returns :class:`requests.Response` instance
        """

        accepted_codes = kwargs.pop('accepted_codes', [100])

        res = self._make_ocs_request(method, service, action, **kwargs)
        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree, accepted_codes=accepted_codes)
            return res

        raise OCSResponseError(res)

    def _make_ocs_request(self, method, service, action, **kwargs):
        """Makes a OCS API request

        :param method: HTTP method
        :param service: service name
        :param action: action path
        :param \*\*kwargs: optional arguments that ``requests.Request.request`` accepts
        :returns :class:`requests.Response` instance
        """
        slash = ''
        if service:
            slash = '/'
        path = self.OCS_BASEPATH + service + slash + action

        attributes = kwargs.copy()

        if 'headers' not in attributes:
            attributes['headers'] = {}

        attributes['headers']['OCS-APIREQUEST'] = 'true'

        if self._debug:
            print('OCS request: %s %s %s' % (method, self.url + path,
                                             attributes))

        res = self._session.request(method, self.url + path, **attributes)
        return res

    def _make_dav_request(self, method, path, **kwargs):
        """Makes a WebDAV request

        :param method: HTTP method
        :param path: remote path of the targetted file
        :param \*\*kwargs: optional arguments that ``requests.Request.request`` accepts
        :returns array of :class:`FileInfo` if the response
        contains it, or True if the operation succeded, False
        if it didn't
        """
        if self._debug:
            print('DAV request: %s %s' % (method, path))
            if kwargs.get('headers'):
                print('Headers: ', kwargs.get('headers'))

        path = self._normalize_path(path)
        res = self._session.request(
            method,
            self._webdav_url + parse.quote(self._encode_string(path)),
            **kwargs
        )
        if self._debug:
            print('DAV status: %i' % res.status_code)
        if res.status_code in [200, 207]:
            return self._parse_dav_response(res)
        if res.status_code in [204, 201]:
            return True
        raise HTTPResponseError(res)

    def _parse_dav_response(self, res):
        """Parses the DAV responses from a multi-status response

        :param res: DAV response
        :returns array of :class:`FileInfo` or False if
        the operation did not succeed
        """
        if res.status_code == 207:
            tree = ET.fromstring(res.content)
            items = []
            for child in tree:
                items.append(self._parse_dav_element(child))
            return items
        return False

    def _parse_dav_element(self, dav_response):
        """Parses a single DAV element

        :param dav_response: DAV response
        :returns :class:`FileInfo`
        """
        href = parse.unquote(
            self._strip_dav_path(dav_response.find('{DAV:}href').text)
        )

        if six.PY2:
            href = href.decode('utf-8')

        file_type = 'file'
        if href[-1] == '/':
            file_type = 'dir'

        file_attrs = {}
        attrs = dav_response.find('{DAV:}propstat')
        attrs = attrs.find('{DAV:}prop')
        for attr in attrs:
            file_attrs[attr.tag] = attr.text

        return FileInfo(href, file_type, file_attrs)

    def _strip_dav_path(self, path):
        """Removes the leading "remote.php/webdav" path from the given path

        :param path: path containing the remote DAV path "remote.php/webdav"
        :returns: path stripped of the remote DAV path
        """
        if path.startswith(self._davpath):
            return path[len(self._davpath):]
        return path

    def _webdav_move_copy(self, remote_path_source, remote_path_target,
                           operation):
        """Copies or moves a remote file or directory

        :param remote_path_source: source file or folder to copy / move
        :param remote_path_target: target file to which to copy / move
        :param operation: MOVE or COPY

        :returns: True if the operation succeeded, False otherwise
        :raises: HTTPResponseError in case an HTTP error status was returned
        """

        if operation != "MOVE" and operation != "COPY":
            return False

        if remote_path_target[-1] == '/':
            remote_path_target += os.path.basename(remote_path_source)

        if not (remote_path_target[0] == '/'):
            remote_path_target = '/' + remote_path_target

        remote_path_source = self._normalize_path(remote_path_source)
        headers = {
            'Destination': self._webdav_url + parse.quote(
                self._encode_string(remote_path_target))
        }

        return self._make_dav_request(
            operation,
            remote_path_source,
            headers=headers
        )

    def _xml_to_dict(self, element):
        """
        Take an XML element, iterate over it and build a dict

        :param element: An xml.etree.ElementTree.Element, or a list of the same
        :returns: A dictionary
        """
        return_dict = {}
        for el in element:
            return_dict[el.tag] = None
            children = list(el)
            if children:
                return_dict[el.tag] = self._xml_to_dict(children)
            else:
                return_dict[el.tag] = el.text
        return return_dict

    def _get_shareinfo(self, data_el):
        """Simple helper which returns instance of ShareInfo class

        :param data_el: 'data' element extracted from _make_ocs_request
        :returns: instance of ShareInfo class
        """
        if (data_el is None) or not (isinstance(data_el, ET.Element)):
            return None
        return ShareInfo(self._xml_to_dict(data_el))

    def _update_capabilities(self):
        res = self._make_ocs_request(
                'GET',
                self.OCS_SERVICE_CLOUD,
                'capabilities'
                )
        if res.status_code == 200:
            tree = ET.fromstring(res.content)
            self._check_ocs_status(tree)

            data_el = tree.find('data')
            apps = {}
            for app_el in data_el.find('capabilities'):
                app_caps = {}
                for cap_el in app_el:
                    app_caps[cap_el.tag] = cap_el.text
                apps[app_el.tag] = app_caps

            self._capabilities = apps

            version_el = data_el.find('version/string')
            edition_el = data_el.find('version/edition')
            self._version = version_el.text
            if edition_el.text is not None:
                self._version += '-' + edition_el.text

            if 'dav' in apps and 'chunking' in apps['dav']:
                chunking_version = float(apps['dav']['chunking'])
                if self._dav_endpoint_version > chunking_version:
                    self._dav_endpoint_version = None

                if self._dav_endpoint_version is None and chunking_version >= 1.0:
                    self._dav_endpoint_version = 1
                else:
                    self._dav_endpoint_version = 0

            return self._capabilities
        raise HTTPResponseError(res)



# -*- coding: utf-8 -*-
#
# vim: expandtab shiftwidth=4 softtabstop=4
#
from setuptools import setup
import io

version = '0.7'

long_description = (
    io.open('README.rst', encoding='utf-8').read()
    + '\n' +
    'Contributors\n'
    '============\n'
    + '\n' +
    io.open('docs/source/CONTRIBUTORS.rst', encoding='utf-8').read()
    + '\n' +
    io.open('CHANGES.rst', encoding='utf-8').read()
    + '\n')

setup(
    name='pyncclient',
    version=version,
    author='Marco Nastasi',
    author_email='m.nastasi@pragmaticminds.de',
    packages=['nextcloud_client', 'nextcloud_client.test'],
    url='https://github.com/pragmaticindustries/pyncclient',
    download_url = 'https://github.com/pragmaticindustries/pyncclient/archive/refs/tags/v0.7.tar.gz',
    license='LICENSE.txt',
    description='Python client library for nextCloud',
    long_description=long_description,
    install_requires=[
        "requests >= 2.0.1",
        "six"
    ],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License'
    ]
)
